# -*- coding: utf_8 -*-
"""
iztools 全场测试功能模拟
"""
import time
from collections import namedtuple
from collections.abc import Callable
from msvcrt import kbhit, getwch
from random import randint
from typing import TypeAlias, Self

from .flow import FlowFactory, TickRunnerResult, FlowManager
from .utils import until, plant_abbr_to_type, zombie_abbr_to_type, randomize_generate_cd
from ..basic.gridstr import parse_grid_str
from ..rp_extend import Controller, HookPosition
from ..structs.game_board import GameBoard, get_board
from ..structs.griditem import Griditem
from ..structs.plant import PlantType, Plant

PlaceZombieOp = namedtuple("PlaceZombieOp", ["type_", "time", "row", "col"])
"""
描述僵尸放置操作的对象

Attributes:
    type_ (ZombieType) : 要放置的僵尸类型
    time (int) : 放置的时间
    row (int) : 放置的行, 从0开始
    col (int) : 放置的列, 从0开始
"""  # Pycharm这里不显示docstring的Attributes

PlantTypeList: TypeAlias = list[list[PlantType | None]]


def parse_plant_type_list(plant_type_str: str) -> tuple[PlantTypeList, PlantTypeList]:
    """
    根据iztools植物字符串生成植物列表

    使用+号表示延迟一轮种植, 即"调控栈位".

    Args:
        plant_type_str: 与izt要求相同的植物列表字符串.
    Returns:
        两个5 * 5列表, 分别表示第一轮, 第二轮种植的植物. 空白值由None(而非PlantType.none)填充
    Raises:
        ValueError: plant_type_string格式错误时抛出
    """
    first_list: list[list[PlantType | None]] = [[None] * 5 for _ in range(5)]
    second_list: list[list[PlantType | None]] = [[None] * 5 for _ in range(5)]
    lines = plant_type_str.strip().splitlines(False)
    if (t := len(lines)) != 5:
        raise ValueError(f"plant_type_string must have 5 lines, instead of {t} lines")
    for row, line in enumerate(lines):
        line = line.strip()
        plus_plant_indices = [i - 1 for i, char in enumerate(line) if char == '+']
        if not plus_plant_indices:
            if (t := len(line)) != 5:
                raise ValueError(f"line {row} must have 5 plants, instead of {t} plants")
            first_list[row] = [plant_abbr_to_type[abbr] for abbr in line]
        else:
            if plus_plant_indices[0] == -1:
                raise ValueError(f"line {row} can't start with +")
            col = 0
            for (i, char) in enumerate(line):
                if char == "+":
                    continue
                if i in plus_plant_indices:
                    second_list[row][col] = plant_abbr_to_type[char]
                else:
                    first_list[row][col] = plant_abbr_to_type[char]
                col += 1
            if col != 5:
                raise ValueError(f"line {row} must have 5 plants, instead of {t} plants")
    return first_list, second_list


def parse_target_list(target_str: str) -> tuple[list[tuple[int, int]], list[int]]:
    """
    根据iztools目标字符串生成目标列表

    Args:
        target_str: 与izt要求相同的目标字符串
    Returns:
        两个列表, 分别表示目标植物和目标脑子的位置
    """
    poses = [parse_grid_str(pos) for pos in target_str.strip().split()]
    return [pos for pos in poses if pos[1] != -1], [pos[0] for pos in poses if pos[1] == -1]


def parse_zombie_place_list(place_zombie_str: str) -> list[PlaceZombieOp]:
    """
    根据iztools僵尸放置字符串生成僵尸放置列表

    Args:
        place_zombie_str: 与izt要求相同的僵尸放置字符串.
    Returns:
        一个列表, 返回所有僵尸操作, 用PlaceZombieOp表示
    Raises:
        ValueError: place_zombie_string格式错误时抛出
    """
    lines = place_zombie_str.strip().splitlines(False)
    if (t := len(lines)) != 3:
        raise ValueError(f"place_zombie_string must have 3 lines, instead of {t} lines")
    types = [zombie_abbr_to_type[abbr] for abbr in lines[0].strip().split()]
    times = [int(time_) for time_ in lines[1].strip().split()]
    rows, cols = zip(*(parse_grid_str(pos) for pos in lines[2].strip().split()))  # zip(*iterable)转置
    if not (len(types) == len(times) == len(rows) == len(cols)):
        raise ValueError("length of types, times, rows and cols must be equal")
    return [PlaceZombieOp(*op) for op in zip(types, times, rows, cols)]


class IzTest:
    """
    模拟iztools全场测试.

    Attributes:
        plant_type_lists: 两个5 * 5列表, 分别表示第一轮, 第二轮种植的植物. 空白值由None(而非PlantType.none)填充.
        place_zombie_list: 一个列表, 表示所有僵尸操作, 用PlaceZombieOp表示.
        repeat_time: 重复次数.
        mj_init_phase: mj初始相位. None表示随机
        target_plants_pos: 目标植物的位置列表. 元素为(row, col)
        target_brains_pos: 目标脑子的位置列表. 元素为row
        game_board: 游戏GameBoard对象.
        controller: 测试使用的Controller对象.
        flow_factory: 生成测试逻辑的FlowFactory对象.
        reset_generate_cd: 是否重置植物的generate_cd, 即, iztools"开启攻击间隔处理"为True
        enable_default_check_end: 是否启用默认的判断输赢功能, 即, 需要手动设置判断时为False
        start_check_end_time: 开始判断一次测试是否输赢的时间, 默认为放下最后一个僵尸的时间.
        end_callback: 一次测试结束时的回调函数, 参数为是否成功bool.
        check_tests_end_callback: 判断是否结束测试的回调函数. 默认为None, 表示按照repeat_time次数重复测试.
            参数为(当前测试次数, 成功次数), 返回None表示不结束, 返回float表示计算概率.
        target_plants: 目标植物列表. 仅在测试时有效, 不建议修改.
        target_brains: 目标脑子列表. 仅在测试时有效, 不建议修改.

    """

    def __init__(self, controller: Controller, reset_generate_cd: bool = True):
        """
        构造IzTest对象

        如此构造的对象不能直接使用! 大部分情况下需要调用init_by_str初始化.

        Args:
            controller: 测试使用的Controller对象.
            reset_generate_cd: 攻击间隔处理 in iztools
        """
        self.plant_type_lists: tuple[PlantTypeList, PlantTypeList] = ([], [])
        self.place_zombie_list: list[PlaceZombieOp] = []
        self.repeat_time: int = 0
        self.mj_init_phase: int | None = None
        self.target_plants_pos: list[tuple[int, int]] = []
        self.target_brains_pos: list[int] = []
        self.game_board: GameBoard = get_board(controller)
        self.controller: Controller = controller
        self.flow_factory: FlowFactory = FlowFactory()
        self.reset_generate_cd: bool = reset_generate_cd
        self.enable_default_check_end: bool = True
        self.start_check_end_time: int = 0
        self.end_callback: Callable[[bool], None] = lambda _: None
        self.check_tests_end_callback: Callable[[int, int], float | None] | None = None

        # 运行时候会时刻改变的量. 不建议修改
        self.target_plants: list[Plant] = []
        self.target_brains: list[Griditem] = []
        self._last_test_ended: bool = False  # 用于判断是否结束一次测试
        self._success_count: int = 0  # 成功次数
        self._test_time: int = 0  # 测试次数

        self._flow_factory_set: bool = False  # 用于判断是否设置了flow_factory

    def init_by_str(self, iztools_str: str) -> Self:
        """
        通过iztools字符串初始化iztest对象

        与iztools的输入格式不完全相同:
            - 允许首尾空行以及每行首尾空格.
            - 支持“测试次数”输入-1表示自定义结束行为, 结束行为默认为测试无限次, 需要self.check_tests_end()手动设置.
            - 支持第二行空行表示无目标: 若此行为空, 则不启用内置的判断输赢功能.
            - 支持不输入8 9 10行表示不放置僵尸: 若此行为空, 则不启用内置的判断输赢功能.
            - (暂且)不支持通过书写顺序调整僵尸编号.

        Args:
            iztools_str: iztools输入字符串
        Returns:
            返回自己
        Raises:
            ValueError: 输入字符串格式错误时抛出
        Examples:
            >>> ctler: Controller = ...
            >>> iz_test = IzTest(ctler).init_by_str('''
            ...     1000 -1
            ...     3-0 4-0 5-0 3-3
            ...     .....
            ...     .....
            ...     bs3_c
            ...     b2ljh
            ...     blyl_
            ...     cg   cg   xg   ww
            ...     0    1    300  700
            ...     4-6  4-6  4-6  4-6''')
            如上为iztools默认例子的输入方式.
        """
        lines = iztools_str.strip().splitlines(False)

        if len(lines) == 7:
            self.place_zombie_list = []
            self.enable_default_check_end = False
        elif len(lines) == 10:
            self.place_zombie_list = parse_zombie_place_list('\n'.join(lines[7:10]))
            self.start_check_end_time = max(op.time for op in self.place_zombie_list)
        else:
            raise ValueError(f"iztools_str must have 7 or 10 lines, not {len(lines)} lines")

        repeat_time, mj_init_phase = map(int, lines[0].strip().split())
        if mj_init_phase < -1 or mj_init_phase >= 460:
            raise ValueError(f"mj_init_phase must be in [-1, 459], not {mj_init_phase}")
        self.mj_init_phase = mj_init_phase if mj_init_phase != -1 else None
        if repeat_time < -1 or repeat_time == 0:
            raise ValueError(f"repeat_time must be positive or -1, not {repeat_time}")
        if repeat_time != -1:
            self.repeat_time = repeat_time
        else:
            self.check_tests_end_callback = lambda _, __: None

        self.target_plants_pos, self.target_brains_pos = parse_target_list(lines[1])
        if self.target_plants_pos == [] and self.target_brains_pos == []:
            self.enable_default_check_end = False

        self.plant_type_lists = parse_plant_type_list('\n'.join(lines[2:7]))
        for target_pos in self.target_plants_pos:
            if (self.plant_type_lists[0][target_pos[0]][target_pos[1]] is None and
                    self.plant_type_lists[1][target_pos[0]][target_pos[1]] is None):
                raise ValueError(f"target plant at {target_pos} is None")
        return self

    def on_game_end(self) -> Callable[[Callable[[bool], None]], Callable[[bool], None]]:
        """
        装饰器, 设置结束时的回调函数

        回调的bool参数为本次测试是否成功.

        Returns:
            添加用装饰器
        """

        def _decorator(func):
            self.end_callback = func
            return func

        return _decorator

    def end(self, succeeded: bool) -> TickRunnerResult:
        """
        返回本函数, 表示本次测试结束

        Args:
            succeeded: 成功则传入True, 失败传入False
        Returns:
            TickRunnerResult.BREAK_RUN
        """
        self._last_test_ended = True
        self.end_callback(succeeded)
        if succeeded:
            self._success_count += 1
        self._test_time += 1

        self.target_plants = []
        self.target_brains = []

        return TickRunnerResult.BREAK_DONE

    def check_tests_end(self) -> Callable[[Callable[[int, int], float | None]], Callable[[int, int], float | None]]:
        """
        装饰器, 设置判断是否结束测试的回调函数

        Returns:
            添加用装饰器
        """

        def _decorator(func):
            self.check_tests_end_callback = func
            return func

        return _decorator

    def set_flow_factory(self,
                         place_priority: int = 10,
                         check_end_priority: int = -10) -> Self:
        """
        设置flow_factory

        Args:
            place_priority: 初始化及放置僵尸tick runner的优先级, 默认为10
            check_end_priority: 判断输赢tick runner的优先级, 默认为-10
        Returns:
            返回自己
        Raises:
            RuntimeError: 已经设置过flow_factory时抛出
        """
        if self._flow_factory_set:
            raise RuntimeError("cannot set flow factory twice!")
        self._flow_factory_set = True

        @self.flow_factory.connect(until(0), only_once=True, priority=place_priority)
        def _init(_):
            # 清掉所有_ObjList的栈
            board = self.game_board
            board.plant_list.free_all().reset_stack()
            board.zombie_list.free_all().reset_stack()
            board.projectile_list.free_all().reset_stack()
            board.griditem_list.free_all().reset_stack()
            for plant_list in self.plant_type_lists:
                for row, line in enumerate(plant_list):
                    for col, type_ in enumerate(line):
                        if type_ is None:
                            continue
                        plant = board.iz_new_plant(row, col, type_)
                        # assert plant is not None
                        if self.reset_generate_cd:
                            randomize_generate_cd(plant)
                        if (row, col) in self.target_plants_pos:
                            self.target_plants.append(plant)
            board.mj_clock = randint(0, 459) if self.mj_init_phase is None else self.mj_init_phase

            for i in range(5):
                brain = self.game_board.new_iz_brain(i)
                if i in self.target_brains_pos:
                    self.target_brains.append(brain)

        for op in self.place_zombie_list:
            @self.flow_factory.add_tick_runner(place_priority)
            def _place_zombie(fm: FlowManager, _op=op):
                if fm.time == _op.time:
                    self.game_board.iz_place_zombie(_op.row, _op.col, _op.type_)
                    return TickRunnerResult.DONE

        if self.enable_default_check_end:
            @self.flow_factory.add_tick_runner(priority=check_end_priority)
            def _check_end(fm: FlowManager):
                if fm.time >= self.start_check_end_time:
                    if (all(plant.is_dead for plant in self.target_plants) and
                            all(brain.id.rank == 0 for brain in self.target_brains)):
                        return self.end(True)  # all iterable对象所有元素为True时候True
                    if self.game_board.zombie_list.obj_num == 0:
                        return self.end(False)
        return self

    def start_test(self, jump_frame: bool = False,
                   speed_rate: float = 1.0,
                   print_interval: int = 10,
                   control_speed_key: str = '\x12') -> tuple[float, float]:
        """
        开始测试

        Args:
            jump_frame: True则开启跳帧测试.
            speed_rate: 速度倍率. 仅当jump_frame = False时有效.
            print_interval: 每隔print_interval次测试打印一次结果. 输入0时代表不打印
            control_speed_key: 非跳帧时切换原速/倍速的按键. 默认值为Ctrl+R
        Returns:
            (测试概率, 使用时间)元组
        """
        start_time = time.time()
        last_time = start_time
        ctler = self.controller
        ctler.open_hook(HookPosition.CHALLENGE_I_ZOMBIE_SCORE_BRAIN)
        if not self._flow_factory_set:
            self.set_flow_factory()
        ctler.start()
        ctler.before()
        if jump_frame:
            ctler.start_jump_frame()
        else:
            frame_duration = 1 if (fd := round(10 / speed_rate)) == 0 else fd
            self.game_board.frame_duration = frame_duration
        ctler.next_frame()

        def __one_test():
            nonlocal last_time
            _flow_manager = self.flow_factory.build_manager()
            ctler.before()
            ctler.next_frame()
            while not self._last_test_ended:
                ctler.before()
                _flow_manager.run()
                if not jump_frame and kbhit() and getwch() == control_speed_key:
                    self.game_board.frame_duration = 10 \
                        if self.game_board.frame_duration != 10 else frame_duration
                ctler.next_frame()
            self._last_test_ended = False
            if print_interval and self._test_time % print_interval == 0:
                print(f"ended {self._test_time} of {self.repeat_time}, "
                      f"success rate: {self._success_count / self._test_time:.2%}, "
                      f"using time: {(t := time.time()) - last_time:.2f}s.")
                last_time = t

        if __callback := self.check_tests_end_callback:
            __one_test()  # no do while!
            while (result := __callback(self._test_time, self._success_count)) is None:
                __one_test()
        else:
            for _ in range(self.repeat_time):
                __one_test()
            result = self._success_count / self.repeat_time

        ctler.before()
        if jump_frame:
            ctler.end_jump_frame()
        else:
            self.game_board.frame_duration = 10
        ctler.close_hook(HookPosition.CHALLENGE_I_ZOMBIE_SCORE_BRAIN)
        ctler.end()
        return result, (time.time() - start_time)



# -*- coding: utf_8 -*-
"""
注入, 打开游戏相关的函数和类.
"""
import os
import subprocess
from typing import overload

from . import asm
from ..rp_extend import Controller
from ..structs.game_board import GameBoard, get_board


def open_game(game_path: str, num: int = 1) -> list[int]:
    """
    通过路径, 将pvz作为python子进程打开游戏
    
    Args:
        game_path: 游戏路径, 绝对相对路径均可
        num: 打开的游戏数量
    Returns:
        打开的所有游戏进程process id组成的列表, 长度为num
    """
    abs_path = os.path.abspath(game_path)
    route, exe_name = os.path.split(abs_path)
    current_directory = os.getcwd()
    os.chdir(route)
    ret = [0] * num
    for i in range(num):
        process = subprocess.Popen(f"\"{exe_name}\"")
        ret[i] = process.pid
    os.chdir(current_directory)
    return ret


def inject(pids: list[int]) -> list[Controller]:
    """
    对pids中的每一个进程注入dll
    
    Args:
        pids: process id列表
    Returns:
        所有进程的Controller对象组成的列表
    """
    current_dir = os.getcwd()
    os.chdir(os.path.dirname(__file__))
    dll_path = os.path.abspath("..\\bin\\rp_dll.dll")
    s = f'..\\bin\\rp_injector.exe \"{dll_path}\" {len(pids)} '
    s += ' '.join([str(i) for i in pids])
    try:
        os.system(s)
    except Exception as e:
        raise e
    finally:
        os.chdir(current_dir)
    return [Controller(pid) for pid in pids]


class InjectedGame:
    """
    描述被注入游戏的类

    Attributes:
        controller: 被注入游戏的控制器
    """
    @overload
    def __init__(self, process_id: int):
        """
        通过process id构造InjectedGame对象

        Args:
            process_id: pvz进程的process id
        """

    @overload
    def __init__(self, game_path: str):
        """
        通过游戏路径构造InjectedGame对象

        Args:
            game_path: pvz主程序路径
        """

    @overload
    def __init__(self, controller: Controller):
        """
        通过Controller对象构造InjectedGame对象

        Args:
            controller: 注入目标游戏的Controller对象
        """

    def __init__(self, arg):
        if isinstance(arg, int):
            self.controller: Controller = Controller(arg)
        elif isinstance(arg, str):
            self.controller: Controller = inject(open_game(arg))[0]
        elif isinstance(arg, Controller):
            self.controller: Controller = arg
        else:
            raise TypeError("the parameter should be int, str or Controller instance")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.controller.end()

    def enter_level(self, level_num: int) -> GameBoard:
        """
        进入游戏, 返回GameBoard对象

        **请切记这个函数会毁坏你原有的关卡存档!**

        Args:
            level_num: 关卡对应数字
        Returns:
            GameBoard对象
        Raises:
            RuntimeError: 若不在载入界面, 主界面或小游戏选项卡界面使用此函数则抛出
        """
        ctler = self.controller
        code = f"""
            push esi;
            mov esi, [{0x6a9ec0}];
            mov eax, [esi + {0x7fc}];
            test eax, eax;
            jz LCompleteLoading;
            cmp eax, 1;
            je LDeleteGameSelector;
            cmp eax, 7;
            je LDeleteChallengeScreen;     
            LError:
            mov [{ctler.result_address}], eax;
            pop esi;
            ret;
            
            LDeleteChallengeScreen:
            call 0x44fd00;  // LawnApp::KillChallengeScreen(esi = LawnApp* this)
            jmp LPreNewGame;
            
            LCompleteLoading:
            mov ecx, esi;
            call {0x452cb0}; // LawnApp::LoadingCompleted(ecx = LawnApp* this)
            
            LDeleteGameSelector:
            call {0x44f9e0}; // LawnApp::KillGameSelector(esi = LawnApp* this)
            
            LPreNewGame:
            push 0;
            push {level_num};
            call 0x44f560;  // LawnApp::PreNewGame
            xor eax, eax;
            mov [{ctler.result_address}], eax;
            pop esi;
            ret;"""  # copied from avz
        while not ctler.read_bool([0x6a9ec0, 0x76c, 0xa1]):  # 是否加载成功bool, thanks for ghast
            continue
        ctler.start()
        ctler.before()
        asm.run(code, ctler)
        ctler.next_frame()
        ctler.before()
        ctler.next_frame()
        ctler.before()
        ctler.end()
        if self.controller.result_i32:
            raise RuntimeError("this function should be used at loading screen, "
                               "main selector screen or challenge selector screen, "
                               f"while the current screen num is {self.controller.result_i32}")
        return get_board(ctler)



# -*- coding: utf_8 -*-
"""
欲速不达前置内容. 撑杆落地平均耗时测试.
"""
from ..flow.flow import FlowManager
from ..flow.iztest import IzTest
from ..rp_extend import Controller
from ..structs.zombie import ZombieStatus


def pole_jumping_test(ctler: Controller, row=8):  # izs对这种记录性多行兼容不是很好. 秀一下.
    t = IzTest(ctler).init_by_str(f"""
        1000 -1
        
        ....o
        ....o
        ....o
        ....o
        ....o
        cg       cg       cg       cg       cg
        0        0        0        0        0
        1-{row}  2-{row}  3-{row}  4-{row}  5-{row} 
        """)

    times = []
    tmp_time = [None] * 5

    @t.flow_factory.add_tick_runner()
    def tag_walking_pvz(fm: FlowManager):
        for idx, z in enumerate(~t.game_board.zombie_list):
            if (not tmp_time[idx]) and z.status == ZombieStatus.pole_vaulting_walking:
                tmp_time[idx] = fm.time

    @t.flow_factory.add_tick_runner()
    def check_end(_):
        nonlocal tmp_time
        if all(tmp_time):
            times.extend(tmp_time)
            tmp_time = [None] * 5
            return t.end(True)

    t.start_test(True, print_interval=100)
    print(sum(times) / len(times), len(times))



# -*- coding: utf_8 -*-
"""
生物钟 脚本示例
"""
from random import randint

from ..flow.iztest import IzTest
from ..flow.utils import until_plant_die, until_plant_last_shoot, repeat, place, until
from ..rp_extend import Controller


def botanical_clock(ctler: Controller, jump_frame=False):  # 生物钟. 卡相位式非定态
    iz_test = IzTest(ctler).init_by_str('''
        1000 -1
        1-2 5-4
        zh_j5
        cptoh
        .....
        3lyp_
        hhwz1
        lz 
        0  
        2-6''')

    @iz_test.flow_factory.add_flow()
    async def place_zombie(_):
        plist = iz_test.game_board.plant_list
        flower = plist["2-5"]
        await until(lambda _: flower.hp <= 4)  # 请不要把plist["2-5"]写在lambda内! 非常影响性能!
        place("cg 2-6")  # 2-5花死前一瞬放撑杆
        star = plist["1-5"]
        await until_plant_last_shoot(star).after(151 - 96)
        # 上面randint不加是必过, 实际上需要判断星星打几下而不是直接找最后一下不攻击再放
        await repeat("xg 1-6")  # 星星最后一发攻击发出后1双鬼
        await until_plant_die(star).after(100)
        await repeat("cg 4-6")  # 星星死后4双杆
        await until_plant_die(plist["4-1"])
        place("cg 5-9")  # 三线死后5-9撑杆
        await until_plant_last_shoot(plist["5-5"]).after(151 + randint(0, 14))
        place("xg 5-6")  # 5-5最后一发攻击发出后5双鬼

    row_one_fail_count = 0
    row_five_fail_count = 0

    @iz_test.on_game_end()
    def end_callback(result: bool):
        if not result:
            nonlocal row_five_fail_count, row_one_fail_count
            plant_list = iz_test.game_board.plant_list
            if plant_list["1-2"] is not None:
                row_one_fail_count += 1
            if plant_list["5-4"] is not None:
                row_five_fail_count += 1

    iz_test.start_test(jump_frame, speed_rate=5)
    print(row_one_fail_count, row_five_fail_count)



# -*- coding: utf_8 -*-
"""
子弹相关的枚举和类
"""
from enum import IntEnum
from typing import Self

from . import obj_base as ob
from ..basic import asm


class ProjectileType(IntEnum):
    pea = 0x0
    snow_pea = 0x1
    cabbage = 0x2
    melon = 0x3
    puff = 0x4
    wintermelon = 0x5
    fire_pea = 0x6
    star = 0x7
    cactus = 0x8
    basketball = 0x9
    kernel = 0xA
    cob_cannon = 0xB
    butter = 0xC


class ProjectileMotionType(IntEnum):
    straight = 0
    parabola = 1
    switch_way = 2
    puff = 5
    left_straight = 6
    starfruit = 7
    cattail = 9


class Projectile(ob.ObjNode):
    ITERATOR_FUNC_ADDRESS = 0x41C9B0

    OBJ_SIZE = 0x94

    int_x = ob.property_i32(0x8, "图像整数x坐标")

    int_y = ob.property_i32(0xc, "图像整数y坐标")

    x = ob.property_f32(0x30, "浮点x坐标")

    y = ob.property_f32(0x34, "浮点y坐标")

    dx = ob.property_f32(0x3c, "x速度")

    dy = ob.property_f32(0x40, "y速度")

    type_ = ob.property_int_enum(0x5c, ProjectileType, "子弹类型")

    motion_type = ob.property_int_enum(0x58, ProjectileMotionType, "子弹运动类型")

    @property
    def target_zombie_id(self) -> ob.ObjId:
        """香蒲子弹目标僵尸"""
        return ob.ObjId(self.base_ptr + 0x88, self._controller)

    def die(self):
        """
        令自己死亡
        """
        code = f"""
            mov eax, {self.base_ptr};
            call {0x46EB20};  // Projectile::Die
            ret;"""
        asm.run(code, self._controller)


class ProjectileList(ob.obj_list(Projectile)):
    def free_all(self) -> Self:
        p_board = self._controller.get_p_board()[1]
        code = f"""
                push edi;   
                push esi;
                mov edi, {p_board}
                mov esi, {self._controller.result_address};
                xor edx, edx;
                mov [esi], edx;
                LIterate:
                    mov {Projectile.ITERATOR_P_BOARD_REG}, edi;
                    call {Projectile.ITERATOR_FUNC_ADDRESS};  // Board::IterateProjectile
                    test al, al;
                    jz LFreeAll;
                    mov eax, [esi]
                    call {0x46EB20};  // Projectile::Die
                    jmp LIterate;
                    
                LFreeAll:
                    mov edi, {self.base_ptr}
                    call {0x41e600};  // DataArray<Zombie>::DataArrayFreeAll
                    pop esi;
                    pop edi;
                    ret;"""
        asm.run(code, self._controller)
        return self



# -*- coding: utf_8 -*-
"""
场地字符串相关的类型与函数
"""
from functools import lru_cache
from typing import TypeAlias

gridstr: TypeAlias = str
"""f'{row}-{col}'字符串, 作为本项目唯一以1开头的行列表示方式"""


@lru_cache()
def parse_grid_str(grid_str: gridstr) -> tuple[int, int]:
    """
    根据f'{row}-{col}'字符串返回(row, col)对象

    Args:
        grid_str: 形如'1-2'的字符串
    Returns:
        (row, col)元组
    """
    return int(grid_str.split('-')[0]) - 1, int(grid_str.split('-')[1]) - 1


def get_grid_str(row: int, col: int) -> gridstr:
    """
    根据row, col返回gridstr
    Args:
        row: 行数
        col: 列数
    Returns:
        返回的gridstr
    """
    return f"{row + 1}-{col + 1}"



# -*- coding: utf_8 -*-
"""
执行编译汇编相关的函数.
"""
from functools import lru_cache

import keystone as ks

from ..rp_extend import Controller


def run(code: str, controller: Controller) -> bool:
    """
    执行code汇编码
    
    Args:
        code: x86 intel格式汇编字符串, 应该是一个(void) -> void函数, 即以"ret"结尾并且保持栈平衡
        controller: Controller对象, 用于执行汇编码
    Returns:
        执行成功返回True
    """
    r = decode(code, controller.asm_address)
    return controller.run_code(r)


__keystone_assembler = ks.Ks(ks.KS_ARCH_X86, ks.KS_MODE_32)


@lru_cache()
def decode(code: str, addr: int = 0) -> bytes:
    """
    解码code汇编码
    
    Args:
        code: x86 intel格式汇编字符串
        addr: 汇编码首字节地址
    Returns:
        解码后的字节码
    """
    asm = __keystone_assembler.asm(code, addr, True)[0]
    return asm



# -*- coding: utf_8 -*-
"""
场地物品相关的枚举和类
"""
from enum import IntEnum
from typing import Self

from . import obj_base as ob
from ..basic import asm


class GriditemType(IntEnum):
    none = 0
    grave = 0x1
    crater = 0x2
    ladder = 0x3
    brain_aq = 0x6
    vase = 0x7
    rake = 0xb
    brain = 0xc


class Griditem(ob.ObjNode):
    """
    场地物品, 包括脑子, 梯子等
    """
    OBJ_SIZE = 0xEC

    ITERATOR_FUNC_ADDRESS = 0x41CAD0

    type_ = ob.property_int_enum(0x8, GriditemType, "场地物品类型")

    row = ob.property_i32(0x14, "所在行")

    col = ob.property_i32(0x10, "所在列")

    brain_hp = ob.property_i32(0x18, """
        脑子血量, 墓碑冒出的量, 弹坑消失倒计时, 钉钯消失倒计时
        
        对于ize中脑子, 初始为70, 每次被啃时-= 1(区别于植物血量 -= 4)
        """)

    layer = ob.property_i32(0x1c, "图层")

    x = ob.property_f32(0x24, "x坐标")

    y = ob.property_f32(0x28, "y坐标")

    def __str__(self):
        return f"#{self.id.index} {self.type_.name} at {self.row + 1}-{self.col + 1}"

    def die(self):
        """
        令自己死亡
        """
        code = f"""
            push esi;
            mov esi, {self.base_ptr};
            call {0x44D000}  // Griditem::GriditemDie
            ret;"""
        asm.run(code, self._controller)


class GriditemList(ob.obj_list(Griditem)):
    def alloc_item(self) -> Griditem:
        """
        从内存数组中申请新的griditem对象

        Returns:
            申请出的Griditem对象
        """
        code = f"""
            push esi;
            mov esi, {self.base_ptr};
            call {0x41E1C0};  // DataArray<GridItem>::DataArrayAlloc
            mov [{self._controller.result_address}], eax;
            pop esi;
            ret;"""
        asm.run(code, self._controller)
        return Griditem(self._controller.result_u32, self._controller)

    def free_all(self) -> Self:
        p_board = self._controller.get_p_board()[1]
        code = f"""
                push ebx;
                push edi;
                push esi;
                mov edi, {p_board};
                mov ebx, {self._controller.result_address}
                mov esi, ebx;
                xor edx, edx;
                mov [esi], edx;
                LIterate:
                    mov {Griditem.ITERATOR_P_BOARD_REG}, edi;
                    call {Griditem.ITERATOR_FUNC_ADDRESS};  // Board::IterateGriditem
                    test al, al;
                    jz LFreeAll;
                    mov esi, [esi];
                    call {0x44D000};  // Griditem::GriditemDie
                    mov esi, ebx;
                    jmp LIterate;
                    
                LFreeAll:
                    mov eax, {self.base_ptr}
                    call {0x41E7D0};  // DataArray<Griditem>::DataArrayFreeAll
                    pop esi;
                    pop edi;
                    pop ebx;
                    ret;"""
        asm.run(code, self._controller)
        return self



# -*- coding: utf_8 -*-
"""
暂停测试过率演示.
"""
from ..flow.iztest import IzTest
from ..rp_extend import Controller


def end_test(ctler: Controller):
    iz_test = IzTest(ctler).init_by_str('''
                 -1 -1
                 2-0
                 .....
                 y_s_b
                 .....
                 .....
                 .....
                 tt
                 0
                 2-6''')

    @iz_test.check_tests_end()
    def end_test_callback(n, ns):  # n: 总次数, ns: 成功次数
        z = 1.96
        diff = z / (ns + z * z) * ((n - ns) * ns / n + z * z / 4) ** 0.5
        if diff > 0.01:
            return None
        return (ns + z * z / 2) / (n + z * z)

    print(iz_test.start_test(True, print_interval=50))



import argparse

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="rpze command line utility")
    parser.add_argument("--path", help="specify a path for running the example")
    args = parser.parse_args()
    if p := args.path:
        print(f"your game path is {p}, "
              f"remember that only 1.0.0.1051_EN on pvz.tools is officially supported")
        try:
            from .basic import InjectedGame
            from .examples.botanical_clock import botanical_clock
        except ImportError as ie:
            raise ImportError("maybe the package is not fully installed?") from ie
        try:
            game = InjectedGame(p)
        except (PermissionError, IOError, FileNotFoundError) as e:
            raise IOError("maybe the path is wrong?") from e
        with game:
            board = game.enter_level(70)
            ctler = game.controller
            botanical_clock(ctler, False)



# -*- coding: utf_8 -*-
"""
僵尸相关的枚举和类
"""
from enum import IntEnum
from typing import Self

from . import obj_base as ob
from ..basic import asm


# 数据结构和pvz_emulator命名保持一致
class ZombieType(IntEnum):
    none = -1
    zombie = 0x0
    flag = 0x1
    conehead = 0x2
    pole_vaulting = 0x3
    buckethead = 0x4
    newspaper = 0x5
    screendoor = 0x6
    football = 0x7
    dancing = 0x8
    backup_dancer = 0x9
    ducky_tube = 0xa
    snorkel = 0xb
    zomboni = 0xc
    dolphin_rider = 0xe
    jack_in_the_box = 0xf
    balloon = 0x10
    digger = 0x11
    pogo = 0x12
    yeti = 0x13
    bungee = 0x14
    ladder = 0x15
    catapult = 0x16
    gargantuar = 0x17
    imp = 0x18
    giga_gargantuar = 0x20


class ZombieStatus(IntEnum):
    walking = 0x0
    dying = 0x1
    dying_from_instant_kill = 0x2
    dying_from_lawnmower = 0x3
    bungee_target_drop = 0x4
    bungee_body_drop = 0x5
    bungee_idle_after_drop = 0x6
    bungee_grab = 0x7
    bungee_raise = 0x8
    bungee_idle = 0xa
    pole_vaulting_running = 0xb
    pole_vaulting_jumping = 0xc
    pole_vaulting_walking = 0xd
    rising_from_ground = 0xe
    jackbox_walking = 0xf
    jackbox_pop = 0x10
    pogo_with_stick = 0x14
    pogo_idle_before_target = 0x15
    pogo_jump_across = 0x1b
    newspaper_walking = 0x1d
    newspaper_destroyed = 0x1e
    newspaper_running = 0x1f
    digger_dig = 0x20
    digger_drill = 0x21
    digger_lost_dig = 0x22
    digger_landing = 0x23
    digger_dizzy = 0x24
    digger_walk_right = 0x25
    digger_walk_left = 0x26
    digger_idle = 0x27
    dancing_moonwalk = 0x28
    dancing_point = 0x29
    dancing_wait_summoning = 0x2a
    dancing_summoning = 0x2b
    dancing_walking = 0x2c
    dancing_armrise1 = 0x2d
    dancing_armrise2 = 0x2e
    dancing_armrise3 = 0x2f
    dancing_armrise4 = 0x30
    dancing_armrise5 = 0x31
    backup_spawning = 0x32
    dolphin_walk_with_dolphin = 0x33
    dophin_jump_in_pool = 0x34
    dolphin_ride = 0x35
    dolphin_jump = 0x36
    dolphin_walk_in_pool = 0x37
    dolphin_walk_without_dolphin = 0x38
    snorkel_walking = 0x39
    snorkel_jump_in_the_pool = 0x3a
    snorkel_swim = 0x3b
    snorkel_up_to_eat = 0x3c
    snorkel_eat = 0x3d
    snorkel_finied_eat = 0x3e
    catapult_shoot = 0x43
    catapult_idle = 0x44
    balloon_flying = 0x49
    balloon_falling = 0x4a
    balloon_walking = 0x4b
    imp_flying = 0x47
    imp_landing = 0x48
    gargantuar_throw = 0x45
    gargantuar_smash = 0x46
    ladder_walking = 0x4c
    ladder_placing = 0x4d
    yeti_escape = 0x5b


class ZombieAction(IntEnum):
    none = 0x0
    entering_pool = 0x1
    leaving_pool = 0x2
    caught_by_kelp = 0x3
    climbing_ladder = 0x6
    falling = 0x7
    fall_from_sky = 0x9


class ZombieAccessoriesType1(IntEnum):
    none = 0x0
    roadcone = 0x1
    bucket = 0x2
    football_cap = 0x3
    miner_hat = 0x4


class ZombieAccessoriesType2(IntEnum):
    none = 0x0
    screen_door = 0x1
    newspaper = 0x2
    ladder = 0x3


class Zombie(ob.ObjNode):
    ITERATOR_FUNC_ADDRESS = 0x41C8F0

    OBJ_SIZE = 0x15c

    int_x = ob.property_i32(0x8, "整数x坐标")

    int_y = ob.property_i32(0xc, "整数y坐标")

    row = ob.property_i32(0x1c, "所在行数")

    type_ = ob.property_int_enum(0x24, ZombieType, "僵尸种类")

    status = ob.property_int_enum(0x28, ZombieStatus, "僵尸状态")

    x = ob.property_f32(0x2c, "浮点x坐标")

    y = ob.property_f32(0x30, "浮点y坐标")

    dx = ob.property_f32(0x34, "x方向速度")

    is_eating = ob.property_bool(0x51, "在啃食时为True")

    flash_cd = ob.property_i32(0x54, """
    发亮倒计时
                                    
    - 刚生成僵尸时为0, 受击变为25
    - 在flash_cd < -500时, 僵尸开始速度重置 + 啃食加速
    """)

    time_since_spawn = ob.property_i32(0x60, "出生时间")

    action = ob.property_int_enum(0x64, ZombieAction, "僵尸行为")

    hp = ob.property_i32(0xc8, "本体血量")

    max_hp = ob.property_u32(0xcc, "本体血量上限")

    accessories_type_1 = ob.property_int_enum(0xc4, ZombieAccessoriesType1, "一类饰品类型")

    accessories_hp_1 = ob.property_i32(0xd0, "一类饰品血量")

    accessories_max_hp_1 = ob.property_i32(0xd4, "一类饰品血量上限")

    accessories_type_2 = ob.property_int_enum(0xd8, ZombieAccessoriesType2, "二类饰品")

    accessories_hp_2 = ob.property_i32(0xdc, "二类饰品血量")

    accessories_max_hp_2 = ob.property_i32(0xe0, "二类饰品血量上限")

    hit_box_x = ob.property_i32(0x8c, "中弹判定横坐标")

    hit_box_y = ob.property_i32(0x90, "中弹判定纵坐标")

    hit_width = ob.property_i32(0x94, "中弹判定宽度")

    hit_height = ob.property_i32(0x98, "中弹判定高度")

    attack_box_x = ob.property_i32(0x9c, "攻击判定横坐标")

    attack_box_y = ob.property_i32(0xa0, "攻击判定纵坐标")

    attack_width = ob.property_i32(0xa4, "攻击判定宽度")

    attack_height = ob.property_i32(0xa8, "攻击判定高度")

    slow_cd = ob.property_i32(0xac, "减速倒计时")

    butter_cd = ob.property_i32(0xb0, "黄油固定倒计时")

    is_dead = ob.property_bool(0xec, '是否"彻底"死亡, 即濒死时此条为False')

    is_not_dying = ob.property_bool(0xba, "不在濒死状态时为True")

    @property
    def master_id(self) -> ob.ObjId:  # 似乎所有ObjNode subclass都用Property而不是Attribute更好看
        """舞王id"""
        return ob.ObjId(self.base_ptr + 0xf0, self._controller)

    @property
    def partner_ids(self) -> tuple[ob.ObjId, ob.ObjId, ob.ObjId, ob.ObjId]:
        """伴舞id"""
        return tuple(ob.ObjId(self.base_ptr + 0xf4 + i * 4, self._controller) for i in range(4))

    def __str__(self) -> str:
        return f"#{self.id.index} {self.type_.name} at row {self.row + 1}"

    def die_no_loot(self):
        """
        令僵尸消失，移除僵尸附件和动画，同时处理除掉落外的僵尸消失相关事件（会触发过关奖品掉落的判定）。
        """
        code = f"""
            mov ecx, {self.base_ptr};
            call {0x530510}; // Zombie::DieNoLoot
            ret;"""
        asm.run(code, self._controller)


class ZombieList(ob.obj_list(Zombie)):
    def free_all(self) -> Self:
        p_board = self._controller.get_p_board()[1]
        code = f"""
            push edi;
            push esi;
            mov edi, {p_board};
            mov esi, {self._controller.result_address};
            xor edx, edx;
            mov [esi], edx;
            LIterate:
                mov {Zombie.ITERATOR_P_BOARD_REG}, edi;
                call {Zombie.ITERATOR_FUNC_ADDRESS};  // Board::IterateZombie
                test al, al;
                jz LFreeAll;
                mov ecx, [esi]
                call {0x530510};  // Zombie::DieNoLoot
                jmp LIterate;
                
            LFreeAll:
                mov edi, {self.base_ptr}
                call {0x41e4d0};  // DataArray<Zombie>::DataArrayFreeAll
                pop esi;
                pop edi;
                ret;"""
        asm.run(code, self._controller)
        return self



# -*- coding: utf_8 -*-
"""
植物相关的枚举和类
"""
import typing
from enum import IntEnum

from . import obj_base as ob
from .obj_base import ObjNode
from ..basic import asm
from ..basic.gridstr import parse_grid_str, gridstr


class PlantType(IntEnum):
    pea_shooter = 0x0
    sunflower = 0x1
    cherry_bomb = 0x2
    wallnut = 0x3
    potato_mine = 0x4
    snow_pea = 0x5
    chomper = 0x6
    repeater = 0x7
    puffshroom = 0x8
    sunshroom = 0x9
    fumeshroom = 0xA
    grave_buster = 0xB
    hypnoshroom = 0xC
    scaredyshroom = 0xD
    iceshroom = 0xE
    doomshroom = 0xF
    lily_pad = 0x10
    squash = 0x11
    threepeater = 0x12
    tangle_kelp = 0x13
    jalapeno = 0x14
    spikeweed = 0x15
    torchwood = 0x16
    tallnut = 0x17
    seashroom = 0x18
    plantern = 0x19
    cactus = 0x1a
    blover = 0x1b
    split_pea = 0x1c
    starfruit = 0x1d
    pumpkin = 0x1e
    magnetshroom = 0x1f
    cabbagepult = 0x20
    flower_pot = 0x21
    kernelpult = 0x22
    coffee_bean = 0x23
    garlic = 0x24
    umbrella_leaf = 0x25
    marigold = 0x26
    melonpult = 0x27
    gatling_pea = 0x28
    twin_sunflower = 0x29
    gloomshroom = 0x2A
    cattail = 0x2B
    winter_melon = 0x2C
    gold_magnet = 0x2D
    spikerock = 0x2E
    cob_cannon = 0x2F
    imitater = 0x30


class PlantStatus(IntEnum):
    idle = 0x0
    wait = 0x1
    work = 0x2
    squash_look = 0x3
    squash_jump_up = 0x4
    squash_stop_in_the_air = 0x5
    squash_jump_down = 0x6
    squash_crushed = 0x7
    grave_buster_land = 0x8
    grave_buster_idle = 0x9
    chomper_bite_begin = 0xA
    chomper_bite_success = 0xB
    chomper_bite_fail = 0xC
    chomper_chew = 0xD
    chomper_swallow = 0xE
    potato_sprout_out = 0xF
    potato_armed = 0x10
    spike_attack = 0x12
    scaredyshroom_scared = 0x14
    scaredyshroom_scared_idle = 0x15
    scaredyshroom_grow = 0x16
    sunshroom_small = 0x17
    sunshroom_grow = 0x18
    sunshroom_big = 0x19
    magnetshroom_working = 0x1A
    magnetshroom_inactive_idle = 0x1B
    cactus_short_idle = 0x1E
    cactus_grow_tall = 0x1F
    cactus_tall_idle = 0x20
    cactus_get_short = 0x21
    tangle_kelp_grab = 0x22
    cob_cannon_unaramed_idle = 0x23
    cob_cannon_charge = 0x24
    cob_cannon_launch = 0x25
    cob_cannon_armed_idle = 0x26
    kernelpult_launch_butter = 0x27
    umbrella_leaf_block = 0x28
    umbrella_leaf_shrink = 0x29
    imitater_explode = 0x2A
    flower_pot_placed = 0x2F
    lily_pad_placed = 0x30


class Plant(ObjNode):
    ITERATOR_FUNC_ADDRESS = 0x41c950

    OBJ_SIZE = 0x14c

    x = ob.property_i32(0x8, "x")

    y = ob.property_i32(0xc, "y")

    visible = ob.property_bool(0x18, "可见时为True")

    row = ob.property_i32(0x1c, "所在行数, 起点为0")

    type_ = ob.property_int_enum(0x24, PlantType, "植物类型")

    col = ob.property_i32(0x28, "int: 所在列数, 起点为0")

    status = ob.property_int_enum(0x3c, PlantStatus, "植物状态")

    hp = ob.property_i32(0x40, "当前血量")

    status_cd = ob.property_i32(0x54, """
        属性倒计时, 如磁铁cd
                                     
        地刺攻击倒计时也在这儿:
            地刺的判断和generate_cd无关. 在范围内有僵尸时使status_cd = 100, == 75时打出攻击
        """)

    generate_cd = ob.property_i32(0x58, """
        子弹生成 / 物品生产倒计时
                                       
        初值为max_boot_delay - 14到max_boot_delay
        """)

    max_boot_delay = ob.property_i32(0x5c, """
        generate_cd的最大值
                                          
        对大多数植物为150，对投手为300，曾为200
        """)

    launch_cd = ob.property_i32(0x90, """
        从准备发射到发射子弹的倒计时
                                     
        **这里有坑, 平常常见的大喷49等数据是两个数据做减法减出来的而不是存在这里的直接数据**
        对于ize常见单发植物, 均在generate_cd == 0时修改launch_cd, launch_cd == 1时候打出子弹.
        其他时候恒为0值.
        ize植物与launch_cd初始数值的关系如下:
            - 豌豆/冰豆/裂荚右: 35
            - 双发/裂荚左 :26
            - 小喷: 29
            - 大喷: 50
            - 杨桃: 40
            - 玉米: 30
            - 胆小: 25
        简单认为攻击所需时间为(上述数值 - 1)即可.
        
        胆小的规律较为复杂:
            - 胆小在launch_cd == 0的时候检测身边僵尸以决定自己是不是缩头
            - 胆小攻击基本规律同上,   同样在== 1时打出子弹
            - 在不是正常站立时, 胆小个人每帧更新generate_cd = 150
        因而, 胆小在常态情况时每帧判断一次周围僵尸决定缩头, 但在攻击前兆时不判断.
        之前零度误认为胆小索敌成功到发射为25也可能源于此, 实际上还是取24更为合适.
        
        对于ize常见双发植物(双发/裂荚左):
            在generate_cd == 25的时候改动一次launch_cd = 26, 即25后打出子弹
            在generate_cd == 0时再改改动一次launch_cd = 26
        """)

    can_attack = ob.property_bool(0x48, "能攻击时为True")

    is_dead = ob.property_bool(0x141, "死亡时为True")

    @property
    def target_zombie_id(self) -> ob.ObjId:
        """倭瓜, 水草目标僵尸编号"""
        return ob.ObjId(self.base_ptr + 0x12c, self._controller)

    def __str__(self) -> str:
        return f"#{self.id.index} {self.type_.name} at {self.row + 1}-{self.col + 1}"

    def die(self):
        """
        令自己死亡
        """
        code = f"""
            push {self.base_ptr};
            call {0x4679b0};  // Plant::Die
            ret;"""
        asm.run(code, self._controller)


class PlantList(ob.obj_list(Plant)):
    def get_by_grid(self, row: int, col: int) -> Plant | None:
        """
        通过row, col找到对应植物
        
        获取编号最小的. 不支持南瓜花盆一类获取所有植物, 更不支持叠种.
        
        Args:
            row: 行数, 从0开始
            col: 列数, 从0开始
        Returns:
            对应位置编号最小的植物, 找不到返回None
        """
        p_board = self._controller.get_p_board()[1]  # 常用函数汇编提速喵
        code = f"""
            push esi;
            push edi;
            mov esi, {self._controller.result_address};
            xor eax, eax;
            mov [esi], eax;
            mov edi, {p_board};

            LIterate:
                mov {Plant.ITERATOR_P_BOARD_REG}, edi;
                call {Plant.ITERATOR_FUNC_ADDRESS};  // Board::IteratePlant
                test al, al;
                jz LNoResult;
                mov eax, [esi];  // eax = Plant*
                cmp dword ptr [eax + {Plant.row.offset}], {row};
                jne LIterate;
                cmp dword ptr [eax + {Plant.col.offset}], {col};
                jne LIterate;
                pop edi;
                pop esi;
                ret;

            LNoResult:
                xor eax, eax;
                mov [esi], eax;
                pop edi;
                pop esi;
                ret;"""
        asm.run(code, self._controller)
        if (result := self._controller.result_u32) != 0:
            return Plant(result, self._controller)
        return None

    @typing.overload
    def __getitem__(self, grid: gridstr) -> Plant | None:
        """
        通过grid_str坐标获取植物

        Args:
            grid: 格子字符串
        Returns:
            对应位置编号最小的植物, 找不到返回None
        """

    def __getitem__(self, item):
        if isinstance(item, str):
            return self.get_by_grid(*parse_grid_str(item))
        return super().__getitem__(item)

    def free_all(self) -> typing.Self:
        p_board = self._controller.get_p_board()[1]
        code = f"""
            push esi;
            push ebx;
            mov ebx, {p_board};
            mov esi, {self._controller.result_address};
            xor edx, edx;
            mov [esi], edx;  // mov [esi], 0 is invalid
            LIterate:
                mov {Plant.ITERATOR_P_BOARD_REG}, ebx;
                call {Plant.ITERATOR_FUNC_ADDRESS};  // Board::IteratePlant
                test al, al;
                jz LFreeAll;
                push dword ptr [esi];
                call {0x4679b0};  // Plant::Die
                jmp LIterate;
                
            LFreeAll:
                mov eax, {self.base_ptr};
                call {0x41E590}; // DataArray<Plant>::DataArrayFreeAll
                pop ebx;
                pop esi;
                ret;"""
        asm.run(code, self._controller)
        return self



# -*- coding: utf_8 -*-
"""
流程控制相关的函数和类
"""

from __future__ import annotations

from collections.abc import Callable, Awaitable, Coroutine, Generator
from enum import Enum, auto
from itertools import count
from typing import TypeAlias, Self


class TickRunnerResult(Enum):
    """TickRunner的返回值"""
    DONE = auto()
    """本tick runner以后再也不执行时返回"""
    BREAK_DONE = auto()  # 不用异常打断. StopIteration形式的返回值和type hint系统匹配程度太差.
    """需要打断本次运行并且以后不再运行本TickRunner时返回"""
    BREAK_ONCE = auto()
    """需要打断本次运行但以后还会运行本TickRunner时返回"""


CondFunc: TypeAlias = Callable[["FlowManager"], bool]
"""判断条件的函数"""
FlowCoroutine: TypeAlias = Coroutine[CondFunc, None, TickRunnerResult | None]
"""Flow返回的协程对象"""
Flow: TypeAlias = Callable[["FlowManager"], FlowCoroutine]
"""await AwaitableCondFunc函数的async def函数"""
TickRunner: TypeAlias = Callable[["FlowManager"], TickRunnerResult | None]
"""帧运行函数, 无返回值表示继续执行, 返回TickRunnerResult表示特殊行为"""
PriorityTickRunner: TypeAlias = tuple[int, TickRunner]
"""带权重的帧运行函数"""


class VariablePool:  # thanks Reisen
    """
    用于表示CondFunc中用于"伴随状态"的默认参数的变量池.
    具体属性由构造函数而定
    """

    def reset(self) -> Self:
        """
        重置变量池. 重置后变量池的属性值为初始值.

        Returns:
            self
        """
        self._args_list = list(self.origin_list)
        for k, v in self.origin_dict.items():
            setattr(self, k, v)
        return self

    def __init__(self, *args, **kwargs):
        """
        Args:
            *args: 匿名变量. 使用下标运算访问.
            **kwargs: 属性. 属性名=初始值
        """
        self._args_list = []
        self.origin_list = args
        self.origin_dict = kwargs
        self.reset()

    def __getitem__(self, item):
        return self._args_list[item]

    def __setitem__(self, key, value):
        self._args_list[key] = value

    def get_all_attrs(self) -> tuple[tuple, dict] | None:
        """
        获取所有属性的值

        Returns:
            所有匿名属性(in a tuple), 所有具名属性(in a dict)
        """
        return tuple(self._args_list), {k: getattr(self, k) for k in self.origin_dict}

    def __str__(self):
        t, d = self.get_all_attrs()
        return f"<{t}, {d}>"


def _await_generator(t):
    yield t


class AwaitableCondFunc(Callable, Awaitable):
    """
    包装CondFunc为Awaitable对象.

    Attributes:
        func: 内层CondFunc函数
    """

    def __init__(self, func: CondFunc):
        self.func: CondFunc = func

    def __call__(self, fm: FlowManager) -> bool:
        """
        调用内层func. 确保AwaitableCondFunc自己也为CondFunc函数.
        """
        return self.func(fm)

    def __await__(self) -> Generator[CondFunc, None, None]:
        """
        让AwaitableCondFunc对象可以await.

        Returns:
            生成器对象. 唯一一个生成结果为self.func.
        """
        return _await_generator(self.func)

    def __and__(self, other: CondFunc) -> Self:
        """
        重载&运算符, 使得对象可以用&运算符, 像逻辑和运算一样连接

        Args:
            other: 另一个CondFunc对象

        Returns:
            一个新的AwaitableCondFunc对象. 该对象的func为self.func and other.func
        """
        return AwaitableCondFunc(lambda fm: self.func(fm) and other(fm))

    def __or__(self, other: CondFunc) -> Self:
        """
        重载|运算符, 使得对象可以用|运算符, 像逻辑或运算一样连接

        Args:
            other: 另一个CondFunc对象
        Returns:
            一个新的AwaitableCondFunc对象. 该对象的func为self.func or other.func
        """
        return AwaitableCondFunc(lambda fm: self.func(fm) or other(fm))

    def __invert__(self) -> Self:
        """
        重载~运算符, 使得对象可以用~运算符, 像逻辑非运算一样.

        Returns:
            一个新的AwaitableCondFunc对象. 该对象的func为not self.func
        """
        return AwaitableCondFunc(lambda fm: not self.func(fm))

    def after(self, delay_time: int) -> Self:
        """
        生成一个 在满足原条件后过delay_time帧返回True的对象.
        Args:
            delay_time: 延迟时间
        Returns:
            一个新的AwaitableCondFunc对象.
        """

        def _cond_func(fm: FlowManager, p=VariablePool(event_time=None)) -> bool:
            if p.event_time is None and self.func(fm):
                p.event_time = fm.time
            if p.event_time is not None and p.event_time + delay_time <= fm.time:
                return True
            return False

        return AwaitableCondFunc(_cond_func)


class FlowManager:
    """
    运行Flow和TickRunner函数的对象

    Attributes:
        tick_runners: 所有TickRunner组成的函数, 运行时按顺序执行
        time: 每执行一次do自增1
    """

    def __init__(self, tick_runners: list[PriorityTickRunner], flows: list[Flow], flow_priority):
        """
        Args:
            tick_runners: tick_runner列表, 以PriorityTickRunner形式提供以便排序
            flows: flow列表
            flow_priority: flows执行优先级
        """
        self._flow_coro_list: list[list] = [[lambda _: True, i(self)] for i in flows]

        def __flow_tick_runner(self_: FlowManager) -> TickRunnerResult | None:
            if not (fcl := self_._flow_coro_list):
                return TickRunnerResult.DONE
            pop_list = []
            for idx, (cond_func, flow) in enumerate(fcl):
                if cond_func(self_):
                    try:
                        fcl[idx][0] = flow.send(None)
                    except StopIteration as se:  # StopIteration.value为返回值
                        pop_list.append(idx)
                        if se.value is None:
                            continue
                        for i in pop_list[::-1]:
                            fcl.pop(i)
                        return se.value
            for i in pop_list[::-1]:
                fcl.pop(i)

        _counter = count()
        tick_runner_list = [(-priority, next(_counter), it) for priority, it in tick_runners]
        tick_runner_list.append((-flow_priority, next(_counter), __flow_tick_runner))
        tick_runner_list.sort()
        # -priority让priority越大优先级别越高
        self.tick_runners: list[TickRunner] = [i[2] for i in tick_runner_list]
        self.time = 0

    def add(self) -> Callable[[TickRunner], TickRunner]:
        """
        运行时添加TickRunner的装饰器

        运行时添加的TickRunner会被放在最后执行. 即, 不支持加优先级, 但确保本帧执行.

        Examples:
            >>> flow_manager: FlowManager = ...
            >>> @flow_manager.add()
            ... def tr(fm: FlowManager) -> TickRunnerResult:
            ...     ...
            为装饰器形式使用

            >>> flow_manager: FlowManager = ...
            >>> def tr(fm: FlowManager) -> TickRunnerResult:
            ...     ...
            >>> flow_manager.add()(tr)
            为函数形式使用
        """

        def _decorator(tr: TickRunner):
            self.tick_runners.append(tr)
            return tr

        return _decorator

    def connect(self, cond: CondFunc, only_once: bool = False) \
            -> Callable[[TickRunner], TickRunner]:
        """
        运行时把tick_runner绑定到cond上的方法, 与add使用方法相同
        
        即 在cond(self)返回true时执行func(self).

        Args:
            cond: 执行func的条件函数. 返回None时按照only_once判断; 返回TickRunnerResult时直接返回.
            only_once: 为True时 只要有一次满足cond则返回
        """

        def _decorator(tr: TickRunner) -> TickRunner:
            def __decorated_tick_runner(fm: FlowManager):
                if cond(fm):
                    ret = tr(fm)
                    if ret is None:
                        return TickRunnerResult.DONE if only_once else None
                    return ret

            self.add()(__decorated_tick_runner)
            return tr

        return _decorator

    def run(self) -> TickRunnerResult | None:
        """
        运行一次内部所有函数

        Returns:
            所有tick_runner都执行完毕时返回DONE, 内部有人打断时返回BREAK_ONCE, 否则返回空.
        """
        if not (trs := self.tick_runners):
            return TickRunnerResult.DONE
        pop_list = []

        def end(_type):
            self.time += 1
            for it in pop_list[::-1]:
                trs.pop(it)
            return _type

        for idx, func in enumerate(trs):
            ret: TickRunnerResult | None = func(self)
            match ret:
                case None:
                    continue
                case TickRunnerResult.DONE:
                    pop_list.append(idx)  # 早该换成链表了
                case TickRunnerResult.BREAK_DONE:
                    pop_list.append(idx)
                    return end(TickRunnerResult.BREAK_ONCE)
                case TickRunnerResult.BREAK_ONCE:
                    return end(TickRunnerResult.BREAK_ONCE)
        return end(None)


class FlowFactory:
    """
    用于生成FlowManager的工厂对象

    Attributes:
        flow_list: 所有flow组成的flow_list
        tick_runner_list: 所有tick_runner组成的列表
    """

    def __init__(self):
        self.flow_list: list[Flow] = []
        self.tick_runner_list: list[PriorityTickRunner] = []

    def add_flow(self) -> Callable[[Flow], Flow]:
        """
        添加flow的方法, 与FlowManager.add使用方法相同
        """

        def _decorator(f: Flow) -> Flow:
            self.flow_list.append(f)
            return f

        return _decorator

    def add_tick_runner(self, priority: int = 0) -> Callable[[TickRunner], TickRunner]:
        """
        添加tick_runner的方法, 与FlowManager.add使用方法相同

        Args:
            priority: 权重 越大越优先执行
        """

        def _decorator(tr: TickRunner):
            self.tick_runner_list.append((priority, tr))
            return tr

        return _decorator

    def connect(self, cond: CondFunc, priority: int = 0, only_once: bool = False) \
            -> Callable[[TickRunner], TickRunner]:
        """
        把tick_runner绑定到cond上的方法, 与FlowManager.add使用方法相同

        Args:
            cond: 执行func的条件函数. 返回None时按照only_once判断; 返回TickRunnerResult时直接返回.
            priority: 权重 越大越优先执行
            only_once: 为true时 仅当第一次满足cond时执行
        """

        def _decorator(tr: TickRunner) -> TickRunner:
            def __decorated_tick_runner(fm: FlowManager):
                if cond(fm):
                    ret = tr(fm)
                    if ret is None:
                        return TickRunnerResult.DONE if only_once else None
                    return ret

            self.add_tick_runner(priority)(__decorated_tick_runner)
            return tr

        return _decorator

    def build_manager(self, flow_priority: int = 0) -> FlowManager:
        """
        生成FlowManager的方法

        Args:
            flow_priority: flow在tick runner中的权重, 越大越优先执行
        Returns:
            生成的FlowManager对象
        """
        return FlowManager(self.tick_runner_list, self.flow_list, flow_priority)



# -*- coding: utf_8 -*- 
from __future__ import annotations

__all__ = ['HookPosition', 'Controller']

from enum import Enum
from typing import Self


class HookPosition(Enum):
    MAIN_LOOP = 0,
    ZOMBIE_PICK_RANDOM_SPEED = 1,
    CHALLENGE_I_ZOMBIE_SCORE_BRAIN = 2
    

class Controller:
    result_bool: bool
    result_f32: float
    result_f64: float
    result_i8: int
    result_i16: int
    result_i32: int
    result_i64: int
    result_u8: int
    result_u16: int
    result_u32: int
    result_u64: int

    @property
    def result_mem(self) -> memoryview: ...
    @property
    def result_address(self) -> int: ...
    @property
    def asm_address(self) -> int: ...
    @property
    def pid(self) -> int: ...

    def __init__(self, pid: int) -> None: ...
    def __eq__(self, other: Self) -> bool: ...
    def __ne__(self, other: Self) -> bool: ...
    def __repr__(self) -> str: ...

    def before(self) -> None: ...
    def next_frame(self) -> None: ...
    def start(self) -> None: ...
    def end(self) -> None: ...
    def start_jump_frame(self) -> bool: ...
    def end_jump_frame(self) -> bool: ...
    def get_p_board(self) -> tuple[bool, int]: ...
    def run_code(self, asm_bytes: bytes) -> bool: ...
    def open_hook(self, hook: HookPosition) -> None: ...
    def close_hook(self, hook: HookPosition) -> None: ...
    def hook_connected(self, hook: HookPosition = HookPosition.MAIN_LOOP) -> bool: ...
    
    def read_bool(self, offsets: list[int]) -> bool | None: ...
    def read_f32(self, offsets: list[int]) -> float | None: ...
    def read_f64(self, offsets: list[int]) -> float | None: ...
    def read_i8(self, offsets: list[int]) -> int | None: ...
    def read_i16(self, offsets: list[int]) -> int | None: ...
    def read_i32(self, offsets: list[int]) -> int | None: ...
    def read_i64(self, offsets: list[int]) -> int | None: ...
    def read_u8(self, offsets: list[int]) -> int | None: ...
    def read_u16(self, offsets: list[int]) -> int | None: ...
    def read_u32(self, offsets: list[int]) -> int | None: ...
    def read_u64(self, offsets: list[int]) -> int | None: ...
    def read_bytes(self, size: int, offsets: list[int]) -> bytes | None: ...
    
    def write_bool(self, value: bool, offsets: list[int]) -> bool: ...
    def write_f32(self, value: float, offsets: list[int]) -> bool: ...
    def write_f64(self, value: float, offsets: list[int]) -> bool: ...
    def write_i8(self, value: int, offsets: list[int]) -> bool: ...
    def write_i16(self, value: int, offsets: list[int]) -> bool: ...
    def write_i32(self, value: int, offsets: list[int]) -> bool: ...
    def write_i64(self, value: int, offsets: list[int]) -> bool: ...
    def write_u8(self, value: int, offsets: list[int]) -> bool: ...
    def write_u16(self, value: int, offsets: list[int]) -> bool: ...
    def write_u32(self, value: int, offsets: list[int]) -> bool: ...
    def write_u64(self, value: int, offsets: list[int]) -> bool: ...
    def write_bytes(self, value: bytes, offsets: list[int]) -> bool: ...


# -*- coding: utf_8 -*-
"""
简化flow编写的工具函数
"""
import random
import typing
from typing import overload

from .flow import FlowManager, AwaitableCondFunc, CondFunc, VariablePool
from ..basic.gridstr import parse_grid_str, gridstr, get_grid_str
from ..structs.game_board import GameBoard, get_board
from ..structs.plant import Plant, PlantType
from ..structs.zombie import ZombieType, Zombie


# flow utils
# 除去delay以外 所有的CondFunc factory以until + 情况命名
@overload
def until(time: int) -> AwaitableCondFunc:
    """
    生成一个 判断时间是否到达 的函数

    Args:
        time: 当前时间大于等于time时返回True
    Examples:
        >>> async def flow(_):
        ...     await until(100)
        ...     ...  # do something
        为一个 在time >= 100时执行do something的flow
    """


@overload
def until(cond_func: CondFunc) -> AwaitableCondFunc:
    """
    把cond_func函数包装为AwaitableCondFunc对象.

    Args:
        cond_func: 判断条件的函数
    Returns:
        一个包装的AwaitableCondFunc函数.
    """


def until(arg):
    if isinstance(arg, int):
        return AwaitableCondFunc(lambda fm: fm.time >= arg)
    return AwaitableCondFunc(arg)


def delay(time: int) -> AwaitableCondFunc:
    """
    生成一个 延迟time帧后返回True 的函数

    Args:
        time: 延迟用时间
    Raises:
        ValueError: time <= 0时候抛出.
    Examples:
        >>> async def flow(_):
        ...     ...  # do something
        ...     place("cg 1-6")
        ...     await delay(50)
        ...     place("cg 2-6")
        ...     ...  # do other thing
        为相隔50cs连放双撑杆
    """
    if time <= 0:
        raise ValueError(f"time must be positive, not {time}")

    def _cond_func(fm: FlowManager, v=VariablePool(start_time=None)) -> bool:
        if v.start_time is None:
            v.start_time = fm.time
        if v.start_time + time - 1 <= fm.time:  # 所有CondFunc函数下一cs开始执行.
            return True
        return False

    return AwaitableCondFunc(_cond_func)


def until_precise_digger(magnetshroom: Plant) -> AwaitableCondFunc:
    """
    生成一个等到磁铁到达精确矿时间的函数

    Args:
        magnetshroom: 要判断cd的磁铁
    Examples:
        >>> async def flow(_):
        ...     magnet: Plant = ...
        ...     ...  # do something
        ...     place("kg 1-6")
        ...     await until_precise_digger(magnet)
        ...     place("kg 2-6")
        ...     ...  # do other thing
        为2-1精确矿
    """
    return AwaitableCondFunc(lambda _: magnetshroom.status_cd <= 587)  # 1500 - 913


def until_plant_die(plant: Plant) -> AwaitableCondFunc:
    """
    生成一个等到植物死亡的函数

    Args:
        plant: 要判断的植物
    """
    return AwaitableCondFunc(lambda _: plant.is_dead)


def until_plant_last_shoot(plant: Plant) -> AwaitableCondFunc:
    """
    生成一个 等到植物"本段最后一次连续攻击" 的函数.

    Args:
        plant: 要判断的植物
    """

    def _cond_func(fm: FlowManager,
                   v=VariablePool(try_to_shoot_time=None, is_shooting_flag=False)):
        if plant.generate_cd == 1:  # 下一帧开打
            v.try_to_shoot_time = fm.time + 1
        if v.try_to_shoot_time == fm.time and plant.launch_cd != 0:  # 在攻击时
            v.is_shooting_flag = True
            return False
        if v.try_to_shoot_time == fm.time and plant.launch_cd == 0:  # 不在攻击时
            t = v.is_shooting_flag
            v.is_shooting_flag = False
            return t  # 上一轮是攻击的 且 这一轮不攻击 返回True
        return False

    return AwaitableCondFunc(_cond_func)


# ize data utils
ize_plant_types: set[PlantType] = {
    PlantType.pea_shooter,
    PlantType.sunflower,
    PlantType.wallnut,
    PlantType.potato_mine,
    PlantType.snow_pea,
    PlantType.chomper,
    PlantType.repeater,
    PlantType.puffshroom,
    PlantType.doomshroom,
    PlantType.scaredyshroom,
    PlantType.squash,
    PlantType.threepeater,
    PlantType.spikeweed,
    PlantType.torchwood,
    PlantType.split_pea,
    PlantType.starfruit,
    PlantType.magnetshroom,
    PlantType.kernelpult,
    PlantType.umbrella_leaf
}
"""所有ize中出现的植物"""

ize_zombie_types: set[ZombieType] = {
    ZombieType.imp,
    ZombieType.conehead,
    ZombieType.pole_vaulting,
    ZombieType.buckethead,
    ZombieType.bungee,
    ZombieType.digger,
    ZombieType.ladder,
    ZombieType.football,
    ZombieType.dancing
}
"""所有ize中出现的僵尸"""

plant_abbr_to_type: dict[str, PlantType | None] = {
    ".": None,
    "1": PlantType.pea_shooter,
    "h": PlantType.sunflower,
    "o": PlantType.wallnut,
    "t": PlantType.potato_mine,
    "b": PlantType.snow_pea,
    "z": PlantType.chomper,
    "2": PlantType.repeater,
    "p": PlantType.puffshroom,
    "d": PlantType.fumeshroom,
    "x": PlantType.scaredyshroom,
    "w": PlantType.squash,
    "3": PlantType.threepeater,
    "_": PlantType.spikeweed,
    "j": PlantType.torchwood,
    "l": PlantType.split_pea,
    "5": PlantType.starfruit,
    "c": PlantType.magnetshroom,
    "y": PlantType.kernelpult,
    "s": PlantType.umbrella_leaf
}
"""植物缩写到植物类型的字典"""

zombie_abbr_to_type: dict[str, ZombieType] = {
    "xg": ZombieType.imp,
    "lz": ZombieType.conehead,
    "cg": ZombieType.pole_vaulting,
    "tt": ZombieType.buckethead,
    "bj": ZombieType.bungee, "xt": ZombieType.bungee,
    "kg": ZombieType.digger,
    "tz": ZombieType.ladder, "ft": ZombieType.ladder,
    "gl": ZombieType.football,
    "ww": ZombieType.dancing, "mj": ZombieType.dancing
}
"""僵尸缩写到僵尸类型的字典"""


# operate utils
def place(place_str: gridstr, board: GameBoard | None = None) -> Zombie | Plant | None:
    """
    用字符串放置植物

    Args:
        place_str: 放置植物/僵尸的字符串
        board: 要放置的board. 为None时使用get_board()
    Returns:
        放置的植物或者僵尸
    Raises:
        ValueError: 无法识别的植物或僵尸缩写
    Examples:
        >>> gb: GameBoard = ...
        >>> place("1 1-2", gb)
        放置一颗豌豆在1-2
        >>> place("cg 1-6")
        放置一个撑杆在1-6
    """
    if board is None:
        board = get_board()
    type_str, pos_str = place_str.split(" ")
    row, col = parse_grid_str(pos_str)
    if (type_ := zombie_abbr_to_type.get(type_str)) is not None:
        return board.iz_place_zombie(row, col, type_)
    if (type_ := plant_abbr_to_type.get(type_str)) is not None:
        return board.iz_new_plant(row, col, type_)
    raise ValueError(f"invalid type_str: {type_str}")


async def repeat(place_str: str,
                 time: int = 2, interval: int = 20, board: GameBoard | None = None):
    """
    生成一个连续放东西的flow

    Args:
        place_str: 放置植物/僵尸的字符串
        time: 放僵尸个数
        interval: 放僵尸间隔时间
        board: 要放置的board. 为None时使用get_board()
    Examples:
        >>> async def flow(_):
        ...    ...  # do something
        ...    await repeat("cg 1-6", time=3)
        为1-6三撑杆
    """
    place(place_str, board)
    for _ in range(time - 1):
        await delay(interval)
        place(place_str, board)


# plant utils
def randomize_generate_cd(plant: Plant) -> Plant:
    """
    令植物的generate_cd按照"放置充分长时间"后的结果随机化

    **仅对can_attack == True植物有效**; 但特判地刺, 地刺王无效.

    具体来说, 其generate_cd概率分布图像为一个梯形:
    上底为max_boot_delay - 14, 下底为max_boot_delay.

    Returns:
        返回传入的植物
    """
    if (not plant.can_attack) or plant.type_ in {PlantType.spikeweed, PlantType.spikerock}:
        return plant
    # 拆成[1, max_ - 14)和[max_ - 14, max_ + 1)两个区间
    # 不可以取0, 可以取max_, max_ - 14和前面概率相等为h
    # h * (max_ - 15) + (h + 0) * 16 / 2 = 1解这个方程, h为梯形的高
    h = 1 / ((max_ := plant.max_boot_delay) - 7)
    distribution = [h] * (max_ - 15) + [h / 15 * i for i in range(15, 0, -1)]
    plant.generate_cd = random.choices(population=range(1, max_ + 1), weights=distribution)[0]
    return plant


@typing.overload
def set_puff_x_offset(puffshroom: Plant, offset: int):
    """
    为小喷设置x偏移

    Args:
        puffshroom: 目标小喷
        offset: 小喷x偏移
    Raises:
        ValueError: offset不在范围内
    Examples:
        >>> p = ...
        >>> set_puff_x_offset(p, 3)
        为小喷设置x偏移为+3
    """


@typing.overload
def set_puff_x_offset(puffshroom: Plant, offsets: typing.Iterable[int]):
    """
    为小喷设置x偏移

    Args:
        puffshroom: 目标小喷
        offsets: 小喷x偏移范围
    Raises:
        ValueError: **最终随机结果的**偏移不在范围内.
    Examples:
        >>> p = ...
        >>> set_puff_x_offset(p, range(-5, 4))
        为小喷设置在整个范围内的随机x偏移
    """


def set_puff_x_offset(puffshroom: Plant, arg):
    center_x = get_board().grid_to_pixel_x(puffshroom.col, puffshroom.row)
    offset = arg if isinstance(arg, int) else random.choice(list(arg))
    if not (-5 <= offset <= 4):
        raise ValueError(f"offset {offset} out of valid range of puffshroom")
    puffshroom.x = center_x + offset


SUNFLOWER_HPS_ON_DROPPING_SUN: tuple[int, int, int, int, int, int, int, int] =\
    tuple(276 - 40 * i for i in range(7)) + (0,)
"""向日葵掉落阳光时的血量"""


def get_sunflower_remaining_sun(sunflower: Plant) -> int:
    """
    获取向日葵剩余阳光

    Args:
        sunflower: 目标向日葵
    Returns:
        向日葵剩余阳光值
    Raises:
        RuntimeError: 出现未知的行为时
    """
    if sunflower.is_dead:
        return 0
    hp = sunflower.hp
    intervals = (300,) + SUNFLOWER_HPS_ON_DROPPING_SUN
    lo, hi = 0, len(intervals) - 1
    while lo <= hi:
        mid = (lo + hi) // 2
        if intervals[mid] >= hp > intervals[mid + 1]:
            return (8 - mid) * 25
        if hp > intervals[mid]:
            hi = mid - 1
        else:
            lo = mid + 1
    raise RuntimeError("unexpected behaviour!")


def get_all_remaining_suns(board: GameBoard | None = None) -> dict[gridstr, int]:
    """
    获取所有向日葵剩余阳光

    Args:
        board: 要获取的board. 为None时使用get_board()
    Returns:
        一个字典, 形式为{向日葵所在的x-y: 剩余阳光值}
    """
    if board is None:
        board = get_board()
    return {get_grid_str(sf.row, sf.col): get_sunflower_remaining_sun(sf)
            for sf in ~board.plant_list if sf.type_ == PlantType.sunflower}



# -*- coding: utf_8 -*-
"""
iztools 默认例子的测试
"""
from ..flow.iztest import IzTest
from ..rp_extend import Controller


def default_test(ctler: Controller, time=1000):  # iztools 默认例子, 兼容性体现
    iz_test = IzTest(ctler).init_by_str(f'''
                 {time} -1
                 3-0 4-0 5-0 3-3
                 .....
                 .....
                 bs3_c
                 b2ljh
                 blyl_
                 cg   cg   xg   ww
                 0    1    300  700
                 4-6  4-6  4-6  4-6''')
    print(iz_test.start_test(True))



from .game_board import get_board, GameBoard
from .plant import PlantType, Plant
from .zombie import ZombieType, Zombie
from .griditem import GriditemType, Griditem
from .projectile import ProjectileType, Projectile



# -*- coding: utf_8 -*-
"""
描述pvz中数据结构的基类和基本函数.
"""
import abc
import collections.abc as c_abc
import typing
from enum import IntEnum

from ..basic import asm
from ..rp_extend import Controller


class ObjBase(abc.ABC):
    """
    pvz中的一个对象

    Attributes:
        base_ptr: 对应pvz中对象的指针
        _controller: 对应pvz的Controller对象
    """

    def __init__(self, base_ptr: int, ctler: Controller):
        """
        一个ObjBase对象由一个指向pvz中的对象的指针, 和对应游戏的Controller构造
        
        Args:
            base_ptr: 游戏中对象的基址
            ctler: 游戏对应的Controller对象
        Raises:
            ValueError: base_ptr为空时抛出
        """
        if base_ptr == 0:
            raise ValueError(f"base_ptr of an {type(self).__name__} object cannot be 0")
        super().__init__()
        self.base_ptr = base_ptr
        self._controller = ctler

    OBJ_SIZE: int = NotImplemented
    """对应pvz类在pvz中的大小, 必须在所有非抽象子类中赋值"""

    def __eq__(self, other: typing.Self) -> bool:
        """
        判断二个ObjBase对象是否指向同一游戏的同一位置

        功能更接近于Python中的is.

        Args:
            other : 另一个ObjBase对象
        """
        return self.base_ptr == other.base_ptr and (self._controller == other._controller)

    def __ne__(self, other: typing.Self) -> bool:
        return not (self.base_ptr == other.base_ptr and (self._controller == other._controller))

    def __str__(self) -> str:
        return (f"<{type(self).__name__} object at [0x{self.base_ptr:x}] "
                f"of process id {self._controller.pid}>")

    def __repr__(self) -> str:
        return (f"{type(self).__name__}(base_ptr=0x{self.base_ptr:x}, "
                f"ctler=Controller({self._controller.pid}))")


class OffsetProperty(property):
    """
    ObjBase对象在pvz内的属性

    Attributes:
        offset: 属性在游戏中的偏移
    """

    def __init__(self, fget, fset, fdel, doc, offset):
        super().__init__(fget, fset, fdel, None)
        self.__doc__ = doc
        self.offset = offset


# property factories 用于生成ObjBase对象在pvz内的属性
def property_bool(offset: int, doc: str):
    def _get(self: ObjBase) -> bool:
        return self._controller.read_bool([self.base_ptr + offset])

    def _set(self: ObjBase, value: bool):
        self._controller.write_bool(value, [self.base_ptr + offset])

    return OffsetProperty(_get, _set, None, "bool: " + doc, offset)


def property_i8(offset: int, doc: str):
    def _get(self: ObjBase) -> int:
        return self._controller.read_i8([self.base_ptr + offset])

    def _set(self: ObjBase, value: int):
        self._controller.write_i8(value, [self.base_ptr + offset])

    return OffsetProperty(_get, _set, None, "int: " + doc, offset)


def property_i16(offset: int, doc: str):
    def _get(self: ObjBase) -> int:
        return self._controller.read_i16([self.base_ptr + offset])

    def _set(self: ObjBase, value: int):
        self._controller.write_i16(value, [self.base_ptr + offset])

    return OffsetProperty(_get, _set, None, "int: " + doc, offset)


def property_i32(offset: int, doc: str):
    def _get(self: ObjBase) -> int:
        return self._controller.read_i32([self.base_ptr + offset])

    def _set(self: ObjBase, value: int):
        self._controller.write_i32(value, [self.base_ptr + offset])

    return OffsetProperty(_get, _set, None, "int: " + doc, offset)


def property_i64(offset: int, doc: str):
    def _get(self: ObjBase) -> int:
        return self._controller.read_i64([self.base_ptr + offset])

    def _set(self: ObjBase, value: int):
        self._controller.write_i64(value, [self.base_ptr + offset])

    return OffsetProperty(_get, _set, None, "int: " + doc, offset)


def property_u8(offset: int, doc: str):
    def _get(self: ObjBase) -> int:
        return self._controller.read_u8([self.base_ptr + offset])

    def _set(self: ObjBase, value: int):
        self._controller.write_u8(value, [self.base_ptr + offset])

    return OffsetProperty(_get, _set, None, "int: " + doc, offset)


def property_u16(offset: int, doc: str):
    def _get(self: ObjBase) -> int:
        return self._controller.read_u16([self.base_ptr + offset])

    def _set(self: ObjBase, value: int):
        self._controller.write_u16(value, [self.base_ptr + offset])

    return OffsetProperty(_get, _set, None, "int: " + doc, offset)


def property_u32(offset: int, doc: str):
    def _get(self: ObjBase) -> int:
        return self._controller.read_u32([self.base_ptr + offset])

    def _set(self: ObjBase, value: int):
        self._controller.write_u32(value, [self.base_ptr + offset])

    return OffsetProperty(_get, _set, None, "int: " + doc, offset)


def property_u64(offset: int, doc: str):
    def _get(self: ObjBase) -> int:
        return self._controller.read_u64([self.base_ptr + offset])

    def _set(self: ObjBase, value: int):
        self._controller.write_u64(value, [self.base_ptr + offset])

    return OffsetProperty(_get, _set, None, "int: " + doc, offset)


def property_f32(offset: int, doc: str):
    def _get(self: ObjBase) -> float:
        return self._controller.read_f32([self.base_ptr + offset])

    def _set(self: ObjBase, value: float):
        self._controller.write_f32(value, [self.base_ptr + offset])

    return OffsetProperty(_get, _set, None, "float: " + doc, offset)


def property_f64(offset: int, doc: str):
    def _get(self: ObjBase) -> float:
        return self._controller.read_f64([self.base_ptr + offset])

    def _set(self: ObjBase, value: float):
        self._controller.write_f64(value, [self.base_ptr + offset])

    return OffsetProperty(_get, _set, None, "float: " + doc, offset)


def property_int_enum(offset: int, cls: type[IntEnum], doc: str):
    def _get(self: ObjBase) -> cls:
        return cls(self._controller.read_i32([self.base_ptr + offset]))

    def _set(self: ObjBase, value: cls):
        self._controller.write_i32(int(value), [self.base_ptr + offset])

    return OffsetProperty(_get, _set, None, f"{cls.__name__}: {doc}", offset)


def property_obj(offset: int, cls: type[ObjBase], doc: str):
    def _get(self: ObjBase) -> cls:
        return cls(self._controller.read_i32([self.base_ptr + offset]), self._controller)

    def _set(self: ObjBase, value: cls):
        if self._controller != value._controller:
            raise ValueError("cannot assign an object from another controller")
        self._controller.write_i32(value.base_ptr, [self.base_ptr + offset])

    return OffsetProperty(_get, _set, None, f"{cls.__name__}: {doc}", offset)


class ObjId(ObjBase):
    """
    ObjNode对象末尾的(index, rank)对象

    游戏内用于ObjNode的识别和储存.
    """

    OBJ_SIZE = 4

    index = property_u16(0, "对象索引")

    rank = property_u16(2, "对象序列号")

    def __eq__(self, val: typing.Self | tuple[int, int]) -> bool:
        """
        ObjId比较相等 与其他ObjId比较或与(index, rank)比较

        Args:
            val: 另一个ObjId对象或(index, rank)一样的可解包对象
        Returns:
            "表示相同对象"返回True
        Raises:
            TypeError: val不是ObjId对象或可解包对象
            ValueError: 可解包不是两个元素
        """
        if isinstance(val, ObjId):
            return ((self._controller.read_u32([self.base_ptr]) ==
                     val._controller.read_u32([val.base_ptr]))
                    and self._controller == val._controller)
        try:
            index, rank = val
        except TypeError as te:
            raise TypeError("ObjId can only compare with another ObjId or"
                            "an unpack-able object like (index, rank), "
                            f"not {type(val).__name__} instance") from te
        except ValueError as ve:
            raise ValueError("unpack-able val should have 2 elements (index, rank)") from ve
        return self._controller.read_u32([self.base_ptr]) == ((rank << 16) | index)

    def __ne__(self, val: typing.Self | tuple[int, int]) -> bool:
        return not (self.__eq__(val))

    def __str__(self) -> str:
        return f"(index={self.index}, rank={self.rank})"


class ObjNode(ObjBase, abc.ABC):
    """
    Plant Zombie等等, 在pvz中由ObjList数组进行内存管理的对象

    Attributes:
        id: ObjNode对象末尾的ObjId对象
    """

    def __init__(self, base_ptr: int, ctler: Controller) -> None:
        super().__init__(base_ptr, ctler)
        self.id = ObjId(base_ptr + self.OBJ_SIZE - 4, ctler)

    ITERATOR_FUNC_ADDRESS: int = NotImplemented
    """返回pvz中迭代对象的函数地址, 必须在所有非抽象子类中赋值"""

    ITERATOR_P_BOARD_REG: str = "edx"
    """迭代对象函数用于存储Board指针的寄存器, reanimation和粒子系统为eax, 其他为edx"""


_T = typing.TypeVar("_T", bound=ObjNode)


class ObjList(ObjBase, c_abc.Sequence[_T], abc.ABC):
    """
    游戏中管理各类对象内存的数组, 即函数表中DataArray对象

    仅帮助type hint用, 请勿直接使用, 而是使用obj_list函数构造.
    """

    OBJ_SIZE = 28

    max_length = property_i32(4, "最大时对象数")

    next_index = property_i32(12, "下一个对象的索引")

    obj_num = property_i32(16, "当前对象数量")

    next_rank = property_i32(20, "下一个对象的序列号")

    def __len__(self) -> int:
        """
        返回最大时对象数, 与max_length相同

        Returns:
            最大时对象数
        """
        return self._controller.read_i32([self.base_ptr + 4])

    def at(self, index: int) -> _T:
        """
        返回index对应下标的元素

        Args:
            index: 非负索引, 不做任何检查
        Returns:
            对应下标的元素, 不确保存活
        """

    @typing.overload
    def __getitem__(self, index: int) -> _T:
        """
        返回index对应下标的元素

        Args:
            index: 整数索引, 即支持负数索引
        Returns:
            对应下标的元素, 不确保存活
        Raises:
            IndexError: 若index超出范围抛出
        """

    @typing.overload
    def __getitem__(self, index: slice) -> list[_T]:
        """
        返回slice切片对应的列表

        若在遍历返回值的时候有新对象生成或死亡, 该方法没法动态调整.

        Args:
            index: 切片索引
        Returns:
            对应切片的列表, 不保证其中任何成员存活
        """

    def __getitem__(self, index): ...

    def __invert__(self) -> c_abc.Iterator[_T]:
        """
        迭代所有未回收对象的迭代器, 利用原版函数在迭代过程中动态寻找下一个对象
        
        Returns:
            迭代器, 仅迭代存活对象
        Examples:
            >>> l: ObjList = ...
            >>> for objects in ~l:
            ...     ...  # do something
            迭代l中所有未回收的对象
        """

    @property
    def alive_iterator(self) -> c_abc.Iterator[_T]:
        """与__invert__()相同"""
        return self.__invert__()

    @typing.overload
    def find(self, index: int | ObjId) -> _T | None:
        """
        通过index查找对象

        用int查找时, 未回收对象返回T.
        用ObjId查找时, 在对应index位置对象rank相同时返回T.

        Args:
            index: 整数索引, ObjId对象或(index, rank)可解包对象
        Returns:
            存在未回收的对应对象返回, 否则返回None.
        Raises:
            TypeError: index不是int和ObjId
        Example:
            >>> self.find(-1)
            当前最后一个对象(len(self) - 1)未回收时返回, 否则返回None
        """

    @typing.overload
    def find(self, idx: int, rank: int) -> _T | None:
        """
        通过(index, rank)组查找对象

        Args:
            idx: 整数索引
            rank: 序列号
        Returns:
            存在未回收的对应对象返回, 否则返回None.
        Example:
            >>> self.find(1, 1)
            若idx == 1的对象的rank==1, 返回该对象, 否则返回None
        """

    def find(self, *args): ...

    def reset_stack(self) -> typing.Self:
        """
        清栈

        在所有对象都被回收时调用. 让本对象之后申请对象从0开始申请

        Returns:
            返回自己
        Raises:
            RuntimeError: 不是所有对象都被回收时候抛出
        """

    @abc.abstractmethod
    def free_all(self) -> typing.Self:  # DataArrayFreeAll会泄露动画对象.
        """
        删除存活的所有对象.

        Returns:
            返回自己
        """


def obj_list(node_cls: type[_T]) -> type[ObjList[_T]]:
    """
    根据node_cls构造对应的NodeClsObject的父类
    
    Args:
        node_cls: ObjNode的子类
    Returns:
        管理node_cls对象的数组的父类
    """

    class _ObjIterator(c_abc.Iterator[_T]):
        def __init__(self, ctler: Controller, _iterate_func_asm):
            self._current_ptr = 0
            self._controller = ctler
            self._iterate_func_asm = _iterate_func_asm

        def __next__(self) -> _T:
            self._controller.result_u64 = self._current_ptr
            self._controller.run_code(self._iterate_func_asm)
            if (self._controller.result_u64 >> 32) == 0:
                raise StopIteration
            self._current_ptr = self._controller.result_u32
            return node_cls(self._current_ptr, self._controller)

        def __iter__(self):
            return self

    class _ObjListImplement(ObjList[_T], abc.ABC):
        def __init__(self, base_ptr: int, ctler: Controller):
            super().__init__(base_ptr, ctler)
            self._array_base_ptr = ctler.read_u32([base_ptr])
            p_board = ctler.get_p_board()[1]
            self._code = f"""
                push esi;
                mov esi, {self._controller.result_address};
                mov {node_cls.ITERATOR_P_BOARD_REG}, {p_board};
                call {node_cls.ITERATOR_FUNC_ADDRESS};
                mov [esi + 4], al;
                pop esi;
                ret;"""  # 可恶的reg优化
            self._iterate_func_asm = None

        def at(self, index: int) -> _T:
            return node_cls(self._array_base_ptr + node_cls.OBJ_SIZE * index, self._controller)

        def find(self, *args) -> _T | None:
            match len(args):
                case 1:
                    index = args[0]
                    if isinstance(index, int):
                        try:
                            target = self[index]
                        except IndexError:
                            return None
                        return target if target.id.rank != 0 else None
                    if isinstance(index, ObjId):
                        target = self.at(index.index)
                        return target if target.id == index else None
                    raise TypeError("index must be int or ObjId instance")
                case 2:
                    idx, rank = args
                    try:
                        target = self[idx]
                    except IndexError:
                        return None
                    return target if target.id.rank == rank else None
                case other:
                    raise ValueError("the function should have 1 or 2 parameters, "
                                     f"not {other} parameters")

        def __getitem__(self, index: int | slice):
            if isinstance(index, int):
                i = index if index >= 0 else index + len(self)
                if i >= len(self) or i < 0:
                    raise IndexError("sequence index out of range")
                return self.at(i)
            if isinstance(index, slice):
                start, stop, step = index.indices(len(self))
                return [self[i] for i in range(start, stop, step)]
            raise TypeError(f"index must be int or slice, not {type(index).__name__} instance")

        def __invert__(self):
            if self._iterate_func_asm is None:
                self._iterate_func_asm = asm.decode(self._code, self._controller.result_address)
            return _ObjIterator(self._controller, self._iterate_func_asm)

        def reset_stack(self):
            if self.obj_num:
                raise RuntimeError(f"cannot reset stack when there are still {self.obj_num} objects alive")
            next_idx = self.next_index
            self.next_index = 0
            length = self.max_length
            self.max_length = 0
            while next_idx != length:
                next_idx = (t := self.at(next_idx).id).index
                t.index = 0
            return self

    return _ObjListImplement



# -*- coding: utf_8 -*-
"""
游戏主界面相关的函数和类
"""
from typing import Self

from . import obj_base as ob
from .griditem import GriditemList, Griditem, GriditemType
from .plant import PlantList, Plant, PlantType
from .projectile import ProjectileList
from ..basic import asm
from ..rp_extend import Controller
from .zombie import ZombieList, ZombieType, Zombie


class GameBoard(ob.ObjBase):
    """
    函数表中Board对象

    增加了一些不属于Board类的方法和属性(大部分为LawnApp和Challenge的), 故命名为GameBoard以示区分.

    Attributes:
        zombie_list: 僵尸列表
        plant_list: 植物列表
        projectile_list: 子弹列表
        griditem_list: 场地物品列表
    """
    OBJ_SIZE = 0x57b0

    def __init__(self, base_ptr: int, controller: Controller):
        super().__init__(base_ptr, controller)
        self.zombie_list: ZombieList = ZombieList(base_ptr + 0x90, controller)
        self.plant_list: PlantList = PlantList(base_ptr + 0xac, controller)
        self.projectile_list: ProjectileList = ProjectileList(base_ptr + 0xc8, controller)
        self.griditem_list: GriditemList = GriditemList(base_ptr + 0x11c, controller)

    _p_challenge = ob.property_u32(0x160, "Challenge对象指针")

    is_dance_mode = ob.property_bool(0x5765, "在dance秘籍时中为True")

    sun_num = ob.property_i32(0x5560, "阳光数量")
    
    game_time = ob.property_i32(0x556c, "游戏时间(包括选卡停留的时间)")

    @property
    def mj_clock(self) -> int:
        """mj时钟"""
        return self._controller.read_i32([0x6a9ec0, 0x838])  # 我真看不懂为什么mj时钟在LawnApp底下啊

    @mj_clock.setter
    def mj_clock(self, value: int):
        self._controller.write_i32(value, [0x6a9ec0, 0x838])

    @property
    def frame_duration(self) -> int:
        """帧时长, 以ms为单位"""
        return self._controller.read_i32([0x6a9ec0, 0x454])

    @frame_duration.setter
    def frame_duration(self, value: int):
        self._controller.write_i32(value, [0x6a9ec0, 0x454])

    def iz_setup_plant(self, plant: Plant):
        """
        对植物进行IZ模式调整, 如纸板化, 土豆出土

        Args:
            plant: 要调整的植物
        Raises:
            ValueError: Challenge对象不存在时抛出
        """
        if not (p_c := self._p_challenge):
            raise ValueError("Challenge object doesn't exist!")
        code = f"""
            mov eax, {p_c};
            push {plant.base_ptr};
            call {0x42A530}; // Challenge::IZombieSetupPlant
            ret;"""
        asm.run(code, self._controller)

    def get_plants_on_lawn(self, row: int, col: int) -> \
            tuple[Plant | None, Plant | None, Plant | None, Plant | None]:
        """
        获取指定格子位置的植物元组. 有(底座, 南瓜, 飞行, 常规)植物四种.

        Args:
            row: 行, 从0开始
            col: 列, 从0开始
        Returns:
            (底座, 南瓜, 飞行, 常规)元组, 对应位置没有植物则返回None.
        """
        code = f"""
            push ebx;
            mov ebx, {(ctler := self._controller).result_address};
            push {row};
            push {col};
            mov edx, {self.base_ptr};
            call {0x40d2a0};
            pop ebx;
            ret;"""
        asm.run(code, ctler)
        view = ctler.result_mem[:16].cast("I")  # 以单个元素为u32格式读取前16字节
        return tuple(None if it == 0 else Plant(it, ctler) for it in view)

    def new_plant(self, row: int, col: int, type_: PlantType) -> Plant:
        """
        关卡内种植植物
        
        此函数不会创建种植的音、特效且不会触发限制种植类关卡种植特定植物的特殊效果.
        
        Args:
            row: 行, 从0开始
            col: 列, 从0开始
            type_: 植物类型
        Returns:
            种植成功的植物对象
        """
        code = f'''
            push -1;
            push {int(type_)};
            push {row};
            push {col};
            mov eax, {self.base_ptr};
            call {0x40CE20};  // Board::NewPlant
            mov [{self._controller.result_address}], eax;
            ret;'''
        asm.run(code, self._controller)
        return Plant(self._controller.result_u32, self._controller)

    def iz_new_plant(self, row: int, col: int, type_: PlantType) -> Plant | None:
        """
        判断植物能否种植在指定格子内, 若能则种植植物并对植物进行我是僵尸关卡的特殊调整.

        Args:
            row: 行数, 0为起点
            col: 列数, 0为起点
            type_: 植物类型
        Returns:
            成功则返回植物对象, 否则返回None
        Raises:
            ValueError: Challenge对象不存在时抛出
        """
        if not (p_c := self._p_challenge):
            raise ValueError("Challenge object doesn't exist!")
        next_idx = self.plant_list.next_index
        code = f"""
            push ebx;
            push edi;
            mov ebx, {row};
            push {col};
            push {int(type_)};
            mov edi, {p_c};
            call {0x42a660}; // Challenge::IZombiePlacePlantInSquare 
            pop edi;
            pop ebx;
            ret;"""
        asm.run(code, self._controller)
        return self.plant_list.find(next_idx)

    def iz_place_zombie(self, row: int, col: int, type_: ZombieType) -> Zombie:
        """
        向指定位置放置僵尸.

        Args:
            row: 行数, 0为起点
            col: 列数, 0为起点
            type_: 僵尸
        Returns:
            放置的僵尸对象
        Raises:
            ValueError: Challenge对象不存在时抛出
        """
        if not (p_c := self._p_challenge):
            raise ValueError("Challenge object doesn't exist!")
        ret_idx = self.zombie_list.next_index
        code = f'''
            mov eax, {row};
            push {col};
            push {int(type_)};
            mov ecx, {p_c};
            call {0x42a0f0};  // Challenge::IZombiePlaceZombie
            ret;'''
        asm.run(code, self._controller)
        return self.zombie_list.at(ret_idx)

    def pixel_to_col(self, x: int, y: int = 0) -> int:
        """
        将坐标转换为列数

        Args:
            x: x坐标
            y: y坐标, 仅在禅境花园有用
        Returns:
            对应的列数, 0开始
        """
        code = f"""
            push edi;
            mov edi, {y};
            mov eax, {x};
            mov ecx, {self.base_ptr};
            call {0x41c4c0};  // Board::PixelToGridX 
            mov [{self._controller.result_address}], eax;
            pop edi;
            ret;"""
        asm.run(code, self._controller)
        return self._controller.result_i32

    def pixel_to_row(self, x: int, y: int) -> int:
        """
        将坐标转换为行数

        Args:
            x: x坐标
            y: y坐标
        Returns:
            对应的行数, 0开始, 对地图外的点, 返回-1
        """
        code = f"""
            mov ecx, {y};
            mov eax, {x};
            mov edx, {self.base_ptr};
            call {0x41c550};  // Board::PixelToGridY
            mov [{self._controller.result_address}], eax;
            ret;"""
        asm.run(code, self._controller)
        return self._controller.result_i32

    def pixel_to_grid(self, x: int, y: int) -> tuple[int, int]:
        """
        将坐标转换为行列

        Args:
            x: x坐标
            y: y坐标
        Returns:
            (row, col)元组， 均从0开始
        """
        return self.pixel_to_row(x, y), self.pixel_to_col(x, y)

    def grid_to_pixel_x(self, col: int, row: int = 0) -> int:
        """
        将行列转换为x坐标

        **请注意参数顺序!!!**

        Args:
            col: 列数, 0开始
            row: 行数, 0开始, 仅在禅境花园有用
        Returns:
            对应的x坐标
        """
        code = f"""
            push esi;
            mov esi, {row};
            mov eax, {col};
            mov ecx, {self.base_ptr};
            call {0x41C680};  // Board::GridToPixelX
            mov [{self._controller.result_address}], eax;
            pop esi;
            ret;"""
        asm.run(code, self._controller)
        return self._controller.result_i32

    def grid_to_pixel_y(self, row: int, col: int) -> int:
        """
        将行列转换为y坐标

        Args:
            row: 行数, 0开始
            col: 列数, 0开始
        Returns:
            对应的y坐标
        """
        code = f"""
            push ebx;
            mov ebx, {self.base_ptr};
            mov eax, {row};
            mov ecx, {col};
            call {0x41c740};  // Board::GridToPixelY
            mov [{self._controller.result_address}], eax;
            pop ebx;
            ret;"""
        asm.run(code, self._controller)
        return self._controller.result_i32

    def grid_to_pixel(self, row: int, col: int) -> tuple[int, int]:
        """
        将行列转换为像素坐标

        Args:
            row: 行数, 0开始
            col: 列数, 0开始
        Returns:
            (x, y)元组
        """
        return self.grid_to_pixel_x(col, row), self.grid_to_pixel_y(row, col)

    def new_iz_brain(self, row: int) -> Griditem:  # 我非常无语为什么这个函数原版没有
        """
        构造一个新的IZ脑子, 用于我是僵尸关卡

        Args:
            row: 构造所在行, 0开始
        Returns:
            构造的IZ脑子
        """
        ret = self.griditem_list.alloc_item()
        ret.type_ = GriditemType.brain
        ret.row = row
        ret.col = 0
        ret.layer = 302000 + 10000 * row  # from cvp
        ret.brain_hp = 70
        ret.x = float(self.grid_to_pixel_x(0, 0) - 40)
        ret.y = float(self.grid_to_pixel_y(row, 0) + 40)
        return ret

    def process_delete_queue(self) -> Self:
        """
        删除所有标记为回收的游戏对象

        Returns:
            返回自己
        """
        code = f"""
            push esi;
            mov esi, {self.base_ptr};
            call {0x41BAD0};  // Board::ProcessDeleteQueue
            pop esi;
            ret;"""
        asm.run(code, self._controller)
        return self


__game_board_cache = None  # 重复构造对象会导致多次decode字节码, 故缓存.


def get_board(controller: Controller | None = None) -> GameBoard:
    """
    获取当前游戏主界面对象

    Args:
        controller: pvz控制器对象. 当为None时, 取得上一个缓存的GameBoard对象.
    Returns:
        当前游戏主界面对象
    Raises:
        RuntimeError: Board对象不存在时抛出
    """
    global __game_board_cache
    if controller is None:
        if __game_board_cache is None:
            raise RuntimeError("Board object doesn't exist!")
        return __game_board_cache
    valid, p_board = controller.get_p_board()
    if not p_board:  # 期待Board对象存在, 用异常不用Optional
        raise RuntimeError("Board object doesn't exist!")
    if (not valid) or (__game_board_cache is None):
        __game_board_cache = GameBoard(p_board, controller)
    return __game_board_cache
